"""Vectice package.

[![PyPI version](/_static/img/current_package.svg)](https://pypi.org/project/vectice/) [![PyPI pyversions](/_static/img/python_versions.svg)](https://pypi.python.org/pypi/vectice/)

The Vectice package is a library allowing data-scientists
to record their progress in the [Vectice app](https://docs.vectice.com/).

This package exposes essential Vectice classes and methods:

- the [connect][vectice.Connection.connect] method
- the [Workspace][vectice.models.Workspace] class
- the [Project][vectice.models.Project] class
- the [Phase][vectice.models.Phase] class
- the [Iteration][vectice.models.Iteration] class
- the [Step][vectice.models.Step] class
- the [Model][vectice.Model] class
"""

from __future__ import annotations

import warnings

warnings.filterwarnings("ignore", ".*PYARROW_IGNORE_TIMEZONE")

from vectice import api, models
from vectice.__version__ import __version__
from vectice.connection import Connection
from vectice.models.dataset import Dataset
from vectice.models.git_version import CodeSource
from vectice.models.model import Metric, Model, Property
from vectice.models.resource import BigQueryResource, FileResource, GCSResource, Resource, S3Resource
from vectice.models.resource.metadata import (
    DatasetSourceOrigin,
    DatasetSourceUsage,
    DatasetType,
    DBMetadata,
    File,
    FilesMetadata,
    MetadataDB,
)
from vectice.utils.logging_utils import _configure_vectice_loggers, disable_logging

connect = Connection.connect

code_capture = True
"""Global code capture flag, enabled by default.

Code capture is triggered when registering a dataset or a model,
and only works when a valid Git repository is found.
Otherwise a warning is logged, telling what might be misconfigured in the repository.

Captured information include the repository name, URL, branch name, commit hash,
and whether the repository is dirty (has uncommitted changes).

Examples:
    To disable code capture globally:

    >>> import vectice
    >>> vectice.code_capture = False

    To re-enable code capture globally:

    >>> import vectice
    >>> vectice.code_capture = True
"""

_configure_vectice_loggers(root_module_name=__name__)
silent = disable_logging
warnings.simplefilter("always", DeprecationWarning)

version = __version__

__all__ = [
    "api",
    "models",
    "version",
    "connect",
    "BigQueryResource",
    "Resource",
    "FileResource",
    "GCSResource",
    "S3Resource",
    "DatasetSourceUsage",
    "DatasetSourceOrigin",
    "DBMetadata",
    "FilesMetadata",
    "File",
    "MetadataDB",
    "DatasetType",
    "silent",
    "Metric",
    "Model",
    "Property",
    "CodeSource",
    "Dataset",
]



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import TYPE_CHECKING

from vectice.api._utils import read_nodejs_date
from vectice.api.json.code import CodeOutput

if TYPE_CHECKING:
    from datetime import datetime


class CodeVersionAction(Enum):
    CREATE_GIT_VERSION = "CREATE_GIT_VERSION"
    CREATE_USER_DECLARED_VERSION = "CREATE_USER_DECLARED_VERSION"


@dataclass
class GitVersion:
    repositoryName: str
    branchName: str
    commitHash: str
    isDirty: bool
    uri: str
    commitComment: str | None = ""
    commitAuthorName: str | None = ""
    commitAuthorEmail: str | None = ""
    entrypoint: str | None = ""


@dataclass
class GitVersionInput:
    repositoryName: str
    branchName: str
    commitHash: str
    isDirty: bool
    uri: str
    commitComment: str | None = None
    commitAuthorName: str | None = None
    commitAuthorEmail: str | None = None
    entrypoint: str | None = None


class UserDeclaredVersion(dict):
    @property
    def name(self) -> str:
        return str(self["name"])

    @property
    def description(self) -> str:
        return str(self["description"])

    @property
    def uri(self) -> str:
        return str(self["uri"])

    @property
    def is_starred(self) -> bool:
        return bool(self["isStarred"])


class CodeVersionCreateBody(dict):
    @property
    def action(self) -> str:
        return str(self["action"])

    @property
    def git_version(self) -> str:
        return str(self["gitVersion"])

    @property
    def user_declared_version(self) -> str:
        return str(self["userDeclaredVersion"])


class CodeVersionInput(dict):
    @property
    def code_id(self) -> int:
        return int(self["codeId"])

    @property
    def name(self) -> str:
        return str(self["name"])

    @property
    def description(self) -> str:
        return str(self["description"])

    @property
    def uri(self) -> str:
        return str(self["uri"])

    @property
    def is_starred(self) -> str:
        return str(self["isStarred"])

    @property
    def git_version_id(self) -> int:
        return int(self["gitVersionId"])

    @property
    def git_version(self) -> GitVersionInput:
        return GitVersionInput(**self["gitVersion"])


class GitVersionOutput(dict):
    @property
    def created_date(self) -> datetime | None:
        return read_nodejs_date(str(self["createdDate"]))

    @property
    def updated_date(self) -> datetime | None:
        return read_nodejs_date(str(self["updatedDate"]))

    @property
    def id(self) -> int:
        return int(self["id"])

    @property
    def version(self) -> int:
        return int(self["version"])

    @property
    def author_id(self) -> int:
        return int(self["authorId"])

    @property
    def workspace_id(self) -> int:
        return int(self["workspaceId"])

    @property
    def deleted_date(self) -> datetime | None:
        return read_nodejs_date(str(self["deletedDate"]))

    @property
    def repository_name(self) -> str:
        return str(self["repositoryName"])

    @property
    def branch_name(self) -> str:
        return str(self["branchName"])

    @property
    def commit_hash(self) -> str:
        return str(self["commitHash"])

    @property
    def commit_comment(self) -> str:
        return str(self["commitComment"])

    @property
    def commit_author_email(self) -> str:
        return str(self["commitAuthorEmail"])

    @property
    def commit_author_name(self) -> str:
        return str(self["commitAuthorName"])

    @property
    def is_dirty(self) -> bool:
        return bool(self["isDirty"])

    @property
    def uri(self) -> str:
        return str(self["uri"])

    @property
    def entrypoint(self) -> str:
        return str(self["entrypoint"])


class CodeVersionOutput(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if "code" in self:
            self._code: CodeOutput = CodeOutput(**self["code"])
        else:
            self._code = None

    def items(self):
        result = []
        for key in self:
            if self[key] is not None:
                result.append((key, self[key]))
        return result

    @property
    def created_date(self) -> datetime | None:
        return read_nodejs_date(str(self["createdDate"]))

    @property
    def updated_date(self) -> datetime | None:
        return read_nodejs_date(str(self["updatedDate"]))

    @property
    def id(self) -> int:
        return int(self["id"])

    @property
    def version(self) -> int:
        return int(self["version"])

    @property
    def name(self) -> str:
        return str(self["name"])

    @property
    def description(self) -> str | None:
        if "description" in self and self["description"] is not None:
            return str(self["description"])
        else:
            return None

    @property
    def uri(self) -> str:
        return str(self["uri"])

    @property
    def author_id(self) -> int:
        return int(self["authorId"])

    @property
    def deleted_date(self) -> datetime | None:
        return read_nodejs_date(str(self["deletedDate"]))

    @property
    def version_number(self) -> int:
        return int(self["versionNumber"])

    @property
    def code_id(self) -> int:
        return int(self["codeId"])

    @property
    def git_version_id(self) -> int:
        return int(self["gitVersionId"])

    @property
    def git_version(self) -> GitVersionOutput | None:
        if "gitVersion" in self:
            return GitVersionOutput(**self["gitVersion"])
        else:
            return None

    @property
    def is_starred(self) -> bool | None:
        if self.get("isStarred"):
            return bool(self["isStarred"])
        else:
            return None

    @property
    def code(self) -> CodeOutput:
        return self._code



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from gql import gql
from gql.transport.exceptions import TransportQueryError

from vectice.api.gql_api import GqlApi, Parser

if TYPE_CHECKING:
    from vectice.api.json.code_version import CodeVersionCreateBody, CodeVersionOutput


_logger = logging.getLogger(__name__)


_RETURNS = """
            createdDate
            updatedDate
            deletedDate
            version
            id
            versionNumber
            name
            description
            uri
            authorId
            codeId
            gitVersionId
            gitVersion {
                        createdDate
                        updatedDate
                        deletedDate
                        version
                        id
                        repositoryName
                        branchName
                        commitHash
                        commitComment
                        commitAuthorEmail
                        commitAuthorName
                        isDirty
                        uri
                        entrypoint
                        authorId
                        workspaceId
                        }
            isStarred
            __typename
            """


class GqlCodeVersionApi(GqlApi):
    def create_code_version(self, code_id: int, code_version: CodeVersionCreateBody):
        variable_types = "$codeId:Float!,$codeVersion:CodeVersionCreationBody!"
        kw = "codeId:$codeId,codeVersion:$codeVersion"
        variables = {"codeId": code_id, "codeVersion": code_version}
        query = GqlApi.build_query(
            gql_query="createCodeVersion",
            variable_types=variable_types,
            returns=_RETURNS,
            keyword_arguments=kw,
            query=False,
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            dataset_output: CodeVersionOutput = Parser().parse_item(response["createCodeVersion"])
            return dataset_output
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "code version", "createCodeVersion")

    def get_code_version(self, code_version: str | int, code_id: int | None = None) -> CodeVersionOutput:
        if isinstance(code_version, int):
            gql_query = "getCodeVersion"
            variable_types = "$codeId:Float!"
            variables = {"codeVersionId": code_version}
            kw = "codeVersionId:$codeVersionId"
        elif isinstance(code_version, str) and code_id:
            gql_query = "getCodeVersionByName"
            variable_types = "$name:String!,$codeId:Float!"
            variables = {"name": code_version, "codeId": code_id}  # type: ignore[dict-item]
            kw = "name:$name,codeId:$codeId"
        else:
            raise ValueError("Missing parameters: string and parent id required.")
        query = GqlApi.build_query(
            gql_query=gql_query, variable_types=variable_types, returns=_RETURNS, keyword_arguments=kw, query=True
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            phase_output: CodeVersionOutput = Parser().parse_item(response[gql_query])
            return phase_output
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "getCodeVersion", code_version)



from __future__ import annotations

import pandas as pd
from pandas import DataFrame, Series

from vectice.models.resource.metadata.column_metadata import (
    DateStat,
)
from vectice.models.resource.metadata.df_wrapper_pandas_default import DataFrameDefaultPandasWrapper


class PandasDFWrapper(DataFrameDefaultPandasWrapper[DataFrame]):
    def __init__(self, dataframe: DataFrame):
        super().__init__(dataframe)

    def __compute_date_column_statistics__(self, series: Series) -> DateStat:
        """Parse a dataframe series and return statistics about it.

        The computed statistics are:
        - the first date
        - the mean
        - the median
        - the last date
        - the count missing value in %
        Parameters:
            series: The pandas series to get information from.

        Returns:
            A DateStat object containing the above statistics.
        """
        # Convert to datetime since mean is not supported for non datetime pandas object such as dbdates
        series = pd.to_datetime(series)
        min_date = series.min().isoformat()
        mean = series.mean().isoformat()
        median = series.median().isoformat()
        max_date = series.max().isoformat()
        missing = series.isnull().sum() / len(series)

        return DateStat(
            missing=float(missing), minimum=str(min_date), mean=str(mean), median=str(median), maximum=str(max_date)
        )



from __future__ import annotations

from enum import Enum

from vectice.models.resource.metadata.dataframe_config import MIN_ROWS_CAPTURE_STATS


class DatasetType(Enum):
    """Enumeration that defines in what shape dataset ares."""

    ORIGIN = "ORIGIN"
    """Raw/origin shape."""
    CLEAN = "CLEAN"
    """Clean shape."""
    VALIDATION = "VALIDATION"
    """Validation shape."""
    MODELING = "MODELING"
    """Modeling shape."""


class DatasetSourceUsage(Enum):
    """Enumeration that defines what datasets are used for."""

    TRAINING = "TRAINING"
    """For training datasets."""
    TESTING = "TESTING"
    """For testing datasets."""
    VALIDATION = "VALIDATION"
    """For validation datasets."""


class DatasetSourceOrigin(Enum):
    """Enumeration that defines where datasets comes from."""

    S3 = "S3"
    """S3 storage."""
    REDSHIFT = "REDSHIFT"
    """Redshift storage."""
    GCS = "GCS"
    """Google Cloud Storage."""
    BIGQUERY = "BIGQUERY"
    """BigQuery storage."""
    SNOWFLAKE = "SNOWFLAKE"
    """Snowflake storage."""
    LOCAL = "LOCAL"
    """Local storage."""
    OTHER = "OTHER"
    """Other storage."""


class DatasetSourceType(Enum):
    """Enumeration that defines the type of datasets."""

    DB = "DB"
    """DB source."""
    FILES = "FILES"
    """Files source."""


class MetadataSettings:
    def __init__(self, minimum_rows_for_statistics: int = MIN_ROWS_CAPTURE_STATS):
        self._minimum_rows_for_statistics = minimum_rows_for_statistics

    @property
    def minimum_rows_for_statistics(self) -> int:
        return self._minimum_rows_for_statistics


class Metadata:
    """This class describes the metadata of a dataset."""

    def __init__(
        self,
        type: DatasetSourceType,
        size: int | None = None,
        usage: DatasetSourceUsage | None = None,
        origin: str | None = None,
    ):
        """Initialize a metadata instance.

        Parameters:
            size: The size of the file.
            type: The type of file.
            usage: The usage made of the data.
            origin: The origin of the data.
        """
        self.size = size
        self.type = type
        self.origin = origin
        self.usage = usage
        self._settings = MetadataSettings()

    def set_settings(self, settings: MetadataSettings):
        self._settings = settings

    def asdict(self) -> dict:
        return {
            "size": self.size,
            "type": self.type.value,
            "usage": self.usage.value if self.usage else None,
            "origin": self.origin,
        }



from __future__ import annotations

from abc import abstractmethod
from typing import Generic, TypeVar

from pandas import Series, api

from vectice.models.resource.metadata.column_metadata import (
    BooleanStat,
    ColumnCategoryType,
    DateStat,
    MostCommon,
    NumericalStat,
    Quantiles,
    Size,
    TextStat,
)
from vectice.models.resource.metadata.dataframe_config import DataFramePandasType
from vectice.models.resource.metadata.df_wrapper_default import DataFrameDefaultWrapper

PT = TypeVar("PT", bound=DataFramePandasType)


class DataFrameDefaultPandasWrapper(DataFrameDefaultWrapper[PT], Generic[PT]):
    rows: int
    columns_numbers: int

    def __init__(self, dataframe: PT):
        super().__init__(dataframe)
        self.rows, self.columns_numbers = self.dataframe.shape

    @abstractmethod
    def __compute_date_column_statistics__(self, series: Series) -> DateStat:
        pass

    def get_size(self) -> Size:
        return Size(rows=int(self.rows), columns=int(self.columns_numbers))

    def is_date_series(self, column: Series) -> bool:
        if api.types.is_datetime64_any_dtype(column):
            return True

        try:
            # Temporary fixing issue -> TypeError: data type 'dbdate' not understood [EN-2534]
            if column.dtypes == "dbdate":
                return True
        except TypeError:
            pass

        return False

    def __capture_column_stats__(
        self, series: Series, minimum_rows_for_statistics: int
    ) -> tuple[ColumnCategoryType | None, TextStat | BooleanStat | NumericalStat | DateStat | None]:
        if api.types.is_bool_dtype(series):
            return (
                ColumnCategoryType.BOOLEAN,
                self.__compute_boolean_column_statistics__(series)
                if self.rows >= minimum_rows_for_statistics
                else None,
            )
        elif api.types.is_numeric_dtype(series):
            return (
                ColumnCategoryType.NUMERICAL,
                self.__compute_numeric_column_statistics__(series)
                if self.rows >= minimum_rows_for_statistics
                else None,
            )
        elif self.is_date_series(series):
            return (
                ColumnCategoryType.DATE,
                self.__compute_date_column_statistics__(series) if self.rows >= minimum_rows_for_statistics else None,
            )
        elif api.types.is_string_dtype(series) | api.types.is_categorical_dtype(series):
            return (
                ColumnCategoryType.TEXT,
                self.__compute_string_column_statistics__(series) if self.rows >= minimum_rows_for_statistics else None,
            )
        return None, None

    def __compute_boolean_column_statistics__(self, series: Series) -> BooleanStat:
        """Parse a dataframe series and return statistics about it.

        The computed statistics are:
        - The percentage of True
        - The percentage of False
        - The count missing value in %
        Parameters:
            series: The pandas series to get information from.

        Returns:
            A BooleanStat object containing the above statistics.
        """
        value_counts = series.value_counts(dropna=False)
        value_counts_percentage = value_counts / value_counts.sum()
        series_missing = series.isnull().sum()
        missing = series_missing / len(series)

        return BooleanStat(
            true=float(value_counts_percentage[True]),
            false=float(value_counts_percentage[False]),
            missing=float(missing),
        )

    def __compute_numeric_column_statistics__(self, series: Series) -> NumericalStat:
        """Parse a dataframe series and return statistics about it.

        The computed statistics are:
        - the mean
        - the standard deviation
        - the min value
        - the 25% percentiles
        - the 50% percentiles
        - the 75% percentiles
        - the max value
        - the count missing value in %
        Parameters:
            series: The pandas series to get information from.

        Returns:
            A NumericalStat object containing the above statistics.
        """
        mean = series.mean()
        std = series.std()
        min_q = series.min()
        q25 = series.quantile(0.25)
        q50 = series.quantile(0.5)
        q75 = series.quantile(0.75)
        max_q = series.max()
        missing = float(series.isnull().sum()) / float(len(series))

        return NumericalStat(
            mean=float(mean),
            std_deviation=float(std),
            quantiles=Quantiles(q_min=float(min_q), q25=float(q25), q50=float(q50), q75=float(q75), q_max=float(max_q)),
            missing=float(missing),
        )

    def __compute_string_column_statistics__(self, series: Series) -> TextStat:
        """Parse a dataframe series and return statistics about it.

        The computed statistics are:
        - the unique number of value
        - the top 3 most common values with their percentages
        - the count missing value in %
        Parameters:
            series: The pandas series to get information from.

        Returns:
            A TextStat object containing the above statistics.
        """
        missing = series.isnull().sum() / len(series)
        unique = len(series.unique())
        value_counts = series.value_counts() / series.value_counts(dropna=False).sum()

        size = 3 if unique >= 3 else unique
        value_counts_largest: Series[float] = value_counts.nlargest(size)

        return TextStat(
            unique=float(unique),
            missing=float(missing),
            most_commons=[
                MostCommon(str(i), float(value_counts_largest[i])) for i in value_counts_largest.index.values
            ],
        )



from vectice.api.json.model_version_representation import ModelVersionRepresentationOutput


class ModelVersionRepresentation:
    def __init__(self, output: ModelVersionRepresentationOutput):
        self.id = output.id
        self.project_id = output.project_id
        self.name = output.name
        self.status = output.status
        self.description = output.description
        self.technique = output.technique
        self.library = output.library



from __future__ import annotations

from vectice.models.resource.metadata.base import (
    DatasetSourceType,
    DatasetSourceUsage,
    Metadata,
)
from vectice.models.resource.metadata.column_metadata import Column
from vectice.models.resource.metadata.dataframe_config import DataFrameType
from vectice.models.resource.metadata.source import Source


class FilesMetadata(Metadata):
    """The metadata of a set of files."""

    def __init__(
        self,
        files: list[File],
        size: int | None = None,
        usage: DatasetSourceUsage | None = None,
        origin: str | None = None,
    ):
        """Initialize a FilesMetadata instance.

        Parameters:
            files: The list of files of the dataset.
            size: The size of the set of files.
            usage: The usage of the dataset.
            origin: Where the dataset files come from.
        """
        super().__init__(size=size, type=DatasetSourceType.FILES, origin=origin, usage=usage)
        self.files = files

    def asdict(self) -> dict:
        for file in self.files:
            if self._settings is not None:
                file.set_settings(self._settings)
        return {
            **super().asdict(),
            "files": [file.asdict() for file in self.files],
        }


class File(Source):
    """Describe a dataset file."""

    def __init__(
        self,
        name: str,
        size: int | None = None,
        fingerprint: str | None = None,
        created_date: str | None = None,
        updated_date: str | None = None,
        uri: str | None = None,
        columns: list[Column] | None = None,
        dataframe: DataFrameType | None = None,
        content_type: str | None = None,
    ):
        """Initialize a file.

        Parameters:
            name: The name of the file.
            size: The size of the file.
            fingerprint: The hash of the file.
            created_date: The date of creation of the file.
            updated_date: The date of last update of the file.
            uri: The uri of the file.
            columns: The columns coming from the dataframe with the statistics.
            dataframe (Optional): A dataframe allowing vectice to optionally compute more metadata about this resource such as columns stats, size, rows number and column numbers. (Support Pandas)
            content_type (Optional): HTTP 'Content-Type' header for this file.
        """
        super().__init__(
            name=name,
            size=size,
            columns=columns,
            created_date=created_date,
            updated_date=updated_date,
            dataframe=dataframe,
        )
        self.fingerprint = fingerprint
        self.uri = uri
        self.content_type = content_type

    def asdict(self) -> dict:
        return {
            **super().asdict(),
            "fingerprint": self.fingerprint,
            "createdDate": self.created_date,
            "uri": self.uri,
            "mimeType": self.content_type,
        }



from __future__ import annotations

import logging
import os
import re
from contextlib import contextmanager
from io import BufferedReader, IOBase
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Union

from gql.transport.exceptions import TransportQueryError
from PIL import Image
from rich.console import Console
from rich.table import Table

from vectice.api.json.iteration import IterationStatus, IterationStepArtifact
from vectice.api.json.step import StepType

if TYPE_CHECKING:
    from vectice.models.iteration import Iteration
    from vectice.models.step import Step


@contextmanager
def hide_logs(package: str):
    old_level = logging.getLogger(package).level
    try:
        logging.getLogger(package).setLevel(logging.ERROR)
        yield
    finally:
        logging.getLogger(package).setLevel(old_level)


def _check_read_only(iteration: Iteration):
    """Check if an iteration is completed or cancelled.

    Refreshing the iteration is necessary because in a Jupyter notebook
    its status could have changed on the backend.

    Parameters:
        iteration: The iteration to check.

    Raises:
        RuntimeError: When the iteration is read-only (completed or cancelled).
    """
    refresh_iteration = iteration._phase.iteration(iteration.index)
    if refresh_iteration._status in {IterationStatus.Completed, IterationStatus.Abandoned}:
        raise RuntimeError(f"The Iteration is {refresh_iteration.status} and is read-only.")


def _get_step_type(
    id: int,
    iteration: Iteration,
    name: str,
    index: int,
    slug: str,
    description: str | None = None,
    artifacts: list[IterationStepArtifact] | None = None,
) -> Step | Any:
    # TODO: cyclic imports
    from vectice.models.step import Step
    from vectice.models.step_dataset import StepDataset
    from vectice.models.step_image import StepImage
    from vectice.models.step_model import StepModel
    from vectice.models.step_number import StepNumber
    from vectice.models.step_string import StepString

    step = Step(
        id=id,
        iteration=iteration,
        name=name,
        index=index,
        slug=slug,
        description=description,
        artifacts=artifacts,
        step_type=StepType.Step,
    )

    def _get_number(text: int | float | str):
        try:
            return float(text)
        except ValueError:
            return text

    if artifacts is not None and len(artifacts) >= 1:
        artifact = artifacts[len(artifacts) - 1]
        if artifact.dataset_version_id is not None:
            return StepDataset(step, artifact)
        if artifact.model_version_id is not None:
            return StepModel(step, artifact)
        if artifact.entity_file_id:
            image = _get_image_info(iteration, artifact.entity_file_id)
            return step if image is None else StepImage(step, image)
        if artifact.text:
            str_or_float = _get_number(artifact.text)
            return StepNumber(step, str_or_float) if isinstance(str_or_float, float) else StepString(step, str_or_float)
    return step


def _get_image_info(iteration: Iteration, entity_file_id: int):
    try:
        image = iteration._client.get_entity_file_by_id(entity_file_id)
        return image.file_name
    except TransportQueryError:
        return None


def _check_image_path(path: str) -> bool:
    try:
        check_path = Path(path).exists()
    except OSError:
        return False
    _, ext = os.path.splitext(path)
    pillow_extensions = {exten for exten in Image.registered_extensions()}
    if not check_path and ext in pillow_extensions:
        raise ValueError("Check the image path.")
    if not check_path and ext not in pillow_extensions:
        return False
    return True


def _validate_image(path: str) -> BufferedReader:
    try:
        return open(path, "rb")
    except FileNotFoundError:
        raise ValueError(f"The provided image {path!r} cannot be opened. Check its format and permissions.") from None


def _get_image_variables(value: str | IOBase | Image.Image) -> tuple[BufferedReader | IOBase, str]:
    if isinstance(value, IOBase):
        image = value
        filename = os.path.basename(value.name)  # type: ignore[attr-defined]
        return image, filename
    if isinstance(value, str):
        image = _validate_image(value)
        filename = os.path.basename(image.name)
        return image, filename
    if isinstance(value, Image.Image):
        image = value.tobytes()
        filename = os.path.basename(value.filename)
        return image, filename

    raise ValueError("Unsupported image provided.")


def _temp_print(string: str | None = None, table: Table | None = None) -> None:
    console = Console(width=120)
    if string:
        print(string)
        print()
    if table:
        console.print(table)
        print()


def _convert_keys_to_camel_case(input_dict: Dict[str, Any]) -> Dict[str, Union[Any, Dict[str, Any]]]:
    camel_case_dict: Dict[str, Union[Any, Dict[str, Any]]] = {}

    for key, value in input_dict.items():
        camel_case_key = re.sub(r"_([a-z])", lambda match: match.group(1).upper(), key)
        if isinstance(value, dict):
            value = _convert_keys_to_camel_case(value)
        camel_case_dict[camel_case_key] = value

    return camel_case_dict



from __future__ import annotations

import logging
import os
from typing import TYPE_CHECKING, Any

from git import GitError
from git.repo import Repo

import vectice
from vectice.api.json.code import CodeInput
from vectice.api.json.code_version import CodeVersionCreateBody, GitVersion
from vectice.utils.common_utils import hide_logs

if TYPE_CHECKING:
    from logging import Logger

    from git.diff import Diff

    from vectice.api.client import Client
    from vectice.api.json.user_declared_version import UserDeclaredVersion


_logger = logging.getLogger(__name__)


class CodeSource:
    """Capture the current git commit."""

    def __init__(self, repository: Repo, user_declared_version: UserDeclaredVersion | None = None):
        self._repository: Repo = repository
        self._git_version: GitVersion | None = _extract_git_version(repository)
        self._user_declared_version: UserDeclaredVersion | None = user_declared_version

    @property
    def repository(self) -> Any:
        return self._repository

    @property
    def user_declared_version(self) -> UserDeclaredVersion | None:
        return self._user_declared_version

    @property
    def git_version(self) -> GitVersion | None:
        return self._git_version


def _check_code_source(client: Client, project_id: str, _logger: Logger) -> int | None:
    """Capture source code.

    Naive implementation that uses the commit hash to name the code.

    Only one version of the code is stored in backend.
    This allows to easily reuse versions based on their commits hash.

    Parameters:
        client: The Vectice client, used to communicate with the backend.
        project_id: The id of the current project.

    Returns:
        Either the id of the created/fetched code version, or none if the code could not be captured.
    """
    repository = _look_for_git_repository()
    if not repository:
        return None
    code = CodeSource(repository)
    if not code.git_version:
        return None
    code_input = CodeInput(name=code.git_version.commitHash)
    try:
        code_output = client.create_code_gql(project_id, code_input)
    except Exception:
        code_output = client.get_code(code.git_version.commitHash, project_id)
        code_version_output = client.get_code_version("Version 1", code_output.id)
        _logger.debug("The code commit exists already.")
        return int(code_version_output.id)
    if code.user_declared_version:
        user_declared_version = code.user_declared_version.__dict__
    else:
        user_declared_version = {}
    code_version_body = CodeVersionCreateBody(
        action="CREATE_GIT_VERSION", gitVersion=code.git_version.__dict__, userDeclaredVersion=user_declared_version
    )
    code_version_output = client.create_code_version_gql(code_output.id, code_version_body)
    _logger.debug("Code captured and will be linked to asset.")
    code_version_id = int(code_version_output.id)
    _capture_local_changed_files(repository, client, project_id, code_version_id)
    return code_version_id


def _capture_local_changed_files(repository: Repo, client: Client, project_id: str, code_version_id: int) -> None:
    diff_outputs = []
    file_names = []
    try:
        changed_files: list[Diff] = repository.index.diff(None)
    except Exception as error:
        changed_files = []
        _logger.warning(
            "Capturing local changed files failed: "
            f"we couldn't get the diff from last commit (detached mode? no commits?): {error!r}"
        )
    for file in changed_files:
        try:
            diff = repository.git.diff(file.a_path)
        except Exception:
            _logger.warning(f"Capturing diff failed for file: {file.a_path}")
            continue
        diff_outputs.append(("file", (f"{file.a_path}.diff", (diff))))
        if file.a_path is not None:
            file_names.append(file.a_path)
    if diff_outputs:
        client.create_code_attachments(diff_outputs, code_version_id, project_id)
    if file_names:
        _logger.info(f"Code captured the following changed files; {', '.join(file_names)}")


def _look_for_git_repository(repo_path=".") -> Any | None:
    try:
        repo_path = os.path.abspath(repo_path)
    except OSError:
        _logger.debug(f"Code capture failed: the directory '{repo_path}' cannot be accessed by the system")
        return None
    try:
        return Repo(repo_path, search_parent_directories=True)
    except GitError as error:
        error_message = str(error) or repo_path
        _logger.debug(
            f"Code capture failed: {error.__class__.__name__}: {error_message}. "
            "Make sure the current directory is a valid Git repository (non-bare, non worktree) "
            "and its permissions allow the current user to access it."
        )
        return None


def _check_for_code(client: Client, project_id: str, _logger: Logger) -> int | None:
    code_version_id = None

    if vectice.code_capture:
        code_version_id = _check_code_source(client, project_id, _logger)
    return code_version_id


def _inform_if_git_repo():
    with hide_logs("vectice"):
        repo = _look_for_git_repository()
    if repo:
        _logger.debug("A git repository was found but code capture is disabled.")


def _extract_git_version(repository: Repo) -> GitVersion | None:
    prefix = "Extracting the Git version failed"
    try:
        url = repository.remotes.origin.url
    except Exception as error:
        _logger.warning(f"{prefix}: we couldn't get the remote URL (no origin?): {error!r}")
        return None
    url = url.rsplit(".git", 1)[0]
    repository_name = url.split("/")[-1]
    try:
        branch_name = repository.active_branch.name
    except Exception as error:
        _logger.warning(f"{prefix}: we couldn't get the branch name (detached mode?): {error!r}")
        return None
    try:
        commit_hash = repository.head.object.hexsha
    except Exception as error:
        _logger.warning(f"{prefix}: we couldn't get the commit hash (no commits?): {error!r})")
        return None
    is_dirty = repository.is_dirty()
    return GitVersion(repository_name, branch_name, commit_hash, is_dirty, url)



from __future__ import annotations

from typing import TYPE_CHECKING

from vectice.api.json.step import StepType
from vectice.models.step import Step

if TYPE_CHECKING:
    from vectice.api.json.iteration import IterationStepArtifact


class StepDataset(Step):
    """Model a dataset step.

    A dataset step stores a dataset version id.
    See also [`Dataset`][vectice.models.dataset.Dataset].
    """

    def __init__(self, step: Step, dataset_version: IterationStepArtifact):
        """Initialize a dataset step.

        Parameters:
            step: The step.
            dataset_version: The step's dataset version.
        """
        super().__init__(
            step.id,
            step._iteration,
            step.name,
            step.index,
            step.slug,
            step._description,
            step_type=StepType.StepDataset,
            artifacts=step.artifacts,
        )
        self._dataset_version = dataset_version

    def __repr__(self):
        return f"StepDataset(name={self.name!r}, id={self.id!r}, description={self._description!r}, dataset_version_id={self.dataset_version!r})"

    @property
    def dataset_version(self) -> IterationStepArtifact:
        """The step's dataset.

        Returns:
            The step's dataset.
        """
        return self._dataset_version



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from gql import gql

from vectice.utils.api_utils import INDEX_ORDERED, PAGINATE_OUTPUT, get_page_input

if TYPE_CHECKING:
    from vectice.api.json import IterationStepArtifactInput, StepOutput

from gql.transport.exceptions import TransportQueryError

from vectice.api.gql_api import GqlApi, Parser
from vectice.api.json.paged_response import PagedResponse
from vectice.api.json.step import StepType

_logger = logging.getLogger(__name__)

_RETURNS_WITH_LIST = """
                    index
                    description
                    slug
                    artifactsCount
                    __typename

            """

_RETURNS_WITH_STEPS = """
                    id
                    index
                    name
                    completed
                    description
                    stepType
                    slug
                    artifacts {
                                entityFileId
                                modelVersion {
                                    vecticeId
                                }
                                datasetVersion {
                                    vecticeId
                                }
                                text
                                type
                    }
                    __typename

            """


_RETURNS = """
                id
                index
                name
                completed
                description
                stepType
                slug
                artifacts {
                                entityFileId
                                modelVersion {
                                    vecticeId
                                }
                                datasetVersion {
                                    vecticeId
                                }
                                text
                                type
                    }
                __typename
            """


_ADD_ARTIFACT_RETURNS = """
                id
                __typename
"""


class StepApi(GqlApi):
    def list_steps(self, iteration_id: str, populate=True) -> PagedResponse[StepOutput]:
        gql_query = "getIterationStepList"
        variable_types = "$parentId:VecticeId!,$order:ListOrderInput,$page:PageInput"
        kw = "parentId:$parentId,order:$order,page:$page"
        variables = {
            "parentId": iteration_id,
            "order": INDEX_ORDERED,
            "page": get_page_input() if not populate else get_page_input(1, 100),
        }
        _returns = _RETURNS_WITH_STEPS if populate is True else _RETURNS_WITH_LIST
        query = GqlApi.build_query(
            gql_query=gql_query,
            variable_types=variable_types,
            returns=PAGINATE_OUTPUT.format(_returns),
            keyword_arguments=kw,
            query=True,
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            step_output: PagedResponse[StepOutput] = Parser().parse_paged_response(response[gql_query])
            return step_output
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "iteration", iteration_id)

    def add_iteration_step_artifact(self, data: IterationStepArtifactInput, step_id: int) -> StepOutput:
        gql_query = "addIterationStepArtifact"
        variable_types = "$id:Float!,$data:IterationStepArtifactInput!"
        variables = {"id": step_id, "data": data}
        kw = "id:$id, data:$data"
        query = GqlApi.build_query(
            gql_query=gql_query,
            variable_types=variable_types,
            returns=_ADD_ARTIFACT_RETURNS,
            keyword_arguments=kw,
            query=False,
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            step_output: StepOutput = Parser().parse_item(response[gql_query])
            return step_output
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "step", step_id)

    def update_iteration_step_artifact(
        self,
        step_id: int,
        step_type: StepType,
        artifacts: list[IterationStepArtifactInput] | None = None,
    ) -> StepOutput:
        gql_query = "updateIterationStepContent"
        variable_types = "$id:Float!,$data:IterationStepUpdateInput!"
        variables: dict = {
            "id": step_id,
            "data": {
                "stepType": step_type.value,
            },
        }
        if artifacts:
            variables["data"]["artifacts"] = artifacts
        kw = "id:$id, data:$data"
        query = GqlApi.build_query(
            gql_query=gql_query,
            variable_types=variable_types,
            returns=_RETURNS,
            keyword_arguments=kw,
            query=False,
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            step_output: StepOutput = Parser().parse_item(response[gql_query])
            return step_output
        except TransportQueryError as e:
            print(e)
            self._error_handler.handle_post_gql_error(e, "step", step_id)

    def get_step(self, step: str, iteration_id: str) -> StepOutput:
        if isinstance(step, str) and iteration_id:
            gql_query = "getStepByName"
            variable_types = "$name:String!,$parentId:VecticeId!"
            variables = {"name": step, "parentId": iteration_id}
            kw = "name:$name,parentId:$parentId"
        else:
            raise ValueError("Missing parameters: string and parent id required.")
        query = GqlApi.build_query(
            gql_query=gql_query, variable_types=variable_types, returns=_RETURNS, keyword_arguments=kw, query=True
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            step_output: StepOutput = Parser().parse_item(response[gql_query])
            return step_output
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "step", step)



from __future__ import annotations

from vectice.api.json.artifact_version import ArtifactVersion, VersionStrategy
from vectice.api.json.attachment import AttachmentOutput
from vectice.api.json.code import CodeInput, CodeOutput
from vectice.api.json.code_version import (
    CodeVersionCreateBody,
    CodeVersionInput,
    CodeVersionOutput,
    GitVersionInput,
    GitVersionOutput,
)
from vectice.api.json.dataset_register import DatasetRegisterOutput
from vectice.api.json.files_metadata import FileMetadata, FileMetadataType
from vectice.api.json.iteration import (
    IterationInput,
    IterationOutput,
    IterationStatus,
    IterationStepArtifact,
    IterationStepArtifactInput,
    IterationStepArtifactType,
)
from vectice.api.json.last_assets import ActivityTargetType, UserActivity
from vectice.api.json.metric import MetricInput, MetricOutput
from vectice.api.json.model_register import ModelRegisterInput, ModelRegisterOutput, ModelType
from vectice.api.json.model_version import ModelVersionInput, ModelVersionOutput, ModelVersionStatus
from vectice.api.json.organization_config import OrgConfigOutput
from vectice.api.json.page import Page
from vectice.api.json.paged_response import PagedResponse
from vectice.api.json.phase import PhaseInput, PhaseOutput
from vectice.api.json.project import ProjectInput, ProjectOutput
from vectice.api.json.property import PropertyInput, PropertyOutput
from vectice.api.json.public_config import ArtifactName, PublicConfigOutput
from vectice.api.json.step import StepInput, StepOutput
from vectice.api.json.user_and_workspace import UserAndDefaultWorkspaceOutput
from vectice.api.json.user_declared_version import UserDeclaredVersion
from vectice.api.json.workspace import WorkspaceInput, WorkspaceOutput

__all__ = [
    "ArtifactVersion",
    "VersionStrategy",
    "GitVersionInput",
    "GitVersionOutput",
    "AttachmentOutput",
    "CodeInput",
    "CodeOutput",
    "CodeVersionInput",
    "CodeVersionOutput",
    "UserAndDefaultWorkspaceOutput",
    "MetricInput",
    "MetricOutput",
    "ModelRegisterInput",
    "ModelRegisterOutput",
    "ModelType",
    "ModelVersionInput",
    "ModelVersionOutput",
    "ModelVersionStatus",
    "UserDeclaredVersion",
    "PagedResponse",
    "ProjectInput",
    "ProjectOutput",
    "PropertyInput",
    "PropertyOutput",
    "FileMetadata",
    "FileMetadataType",
    "WorkspaceOutput",
    "WorkspaceInput",
    "Page",
    "PhaseInput",
    "PhaseOutput",
    "StepInput",
    "StepOutput",
    "IterationInput",
    "IterationOutput",
    "IterationStatus",
    "IterationStepArtifactInput",
    "IterationStepArtifact",
    "IterationStepArtifactType",
    "DatasetRegisterOutput",
    "ModelRegisterOutput",
    "ActivityTargetType",
    "UserActivity",
    "CodeVersionCreateBody",
    "ArtifactName",
    "PublicConfigOutput",
    "OrgConfigOutput",
]



from __future__ import annotations

import logging
from abc import abstractmethod
from typing import Generic, TypeVar

from pandas import Series
from pyspark.sql.dataframe import DataFrame as SparkDF

from vectice.models.resource.metadata.column_metadata import Column, Size
from vectice.models.resource.metadata.dataframe_config import (
    MAX_COLUMNS_CAPTURE_STATS,
    DataFrameTypeWithoutWrapper,
)
from vectice.models.resource.metadata.df_wrapper_resource import DataFrameWrapper

_logger = logging.getLogger(__name__)


DT = TypeVar("DT", bound=DataFrameTypeWithoutWrapper)


class DataFrameDefaultWrapper(DataFrameWrapper[DT], Generic[DT]):
    def __init__(self, dataframe: DT):
        super().__init__(dataframe)

    @abstractmethod
    def get_size(self) -> Size:
        pass

    @abstractmethod
    def __capture_column_stats__(self, column, minimum_rows_for_statistics: int):
        pass

    def capture_columns(self, minimum_rows_for_statistics: int) -> list[Column]:
        columns: list[Column] = []
        dtypes = self.dataframe.dtypes
        column_names_with_types = dtypes.astype(str).to_dict().items() if isinstance(dtypes, Series) else dtypes
        for idx, (name, d_type) in enumerate(column_names_with_types):
            if idx >= MAX_COLUMNS_CAPTURE_STATS:
                _logger.warning(
                    f"Statistics are only captured for the first {MAX_COLUMNS_CAPTURE_STATS} columns of your dataframe."
                )
                break
            column = self.dataframe.select(name) if isinstance(self.dataframe, SparkDF) else self.dataframe[name]
            category_type, stats = self.__capture_column_stats__(
                column, minimum_rows_for_statistics=minimum_rows_for_statistics
            )
            columns.append(
                Column(
                    name=name,
                    data_type=d_type if d_type != "object" else "string",
                    stats=stats,
                    category_type=category_type,
                )
            )

        return columns



from __future__ import annotations

import logging
import re
from typing import TYPE_CHECKING

from gql import gql
from gql.transport.exceptions import TransportQueryError

from vectice.api.gql_api import GqlApi, Parser
from vectice.utils.api_utils import INDEX_ORDERED, PAGINATE_OUTPUT, get_page_input
from vectice.utils.vectice_ids_regex import PHASE_VID_REG

if TYPE_CHECKING:
    from vectice.api.json.paged_response import PagedResponse
    from vectice.api.json.phase import PhaseOutput
    from vectice.api.json.step import StepOutput


_RETURNS = """vecticeId
              name
              status
              index
              owner {
                name
              }
              __typename
            """

_RETURNS_LIST = f"""
    {_RETURNS}
    iterationsCount {{
        notStarted
        inProgress
        inReview
        total
    }}
    stepsCount
"""

_RETURNS_PAGE = PAGINATE_OUTPUT.format(_RETURNS_LIST)

_PARENT_FULL = """
                parent {
                    vecticeId
                    name
                    description
                    workspace {
                        vecticeId
                        name
                        description
                        __typename
                    }
                    __typename
                }
"""

_RETURNS_FULL = f"""
    {_RETURNS}
    {_PARENT_FULL}
"""

_STEP_DEFINITION_RETURNS = """id
              name
              index
              slug
              description
              __typename
            """


_logger = logging.getLogger(__name__)


class PhaseApi(GqlApi):
    def list_phases(self, parent_id: str) -> PagedResponse[PhaseOutput]:
        gql_query = "getPhaseList"

        variable_types = "$filters:PhaseFiltersInput!,$order:ListOrderInput,$page:PageInput"
        kw = "filters:$filters,order:$order,page:$page"
        variables = {
            "filters": {
                "parentId": parent_id,
            },
            "order": INDEX_ORDERED,
            "page": get_page_input(),
        }
        query = GqlApi.build_query(
            gql_query=gql_query, variable_types=variable_types, returns=_RETURNS_PAGE, keyword_arguments=kw
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            phase_output: PagedResponse[PhaseOutput] = Parser().parse_paged_response(response[gql_query])
            return phase_output
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "phase", "list")

    def get_phase(self, phase: str, parent_id: str | None = None, full: bool = False) -> PhaseOutput:
        if re.search(PHASE_VID_REG, phase):
            gql_query = "getPhaseById"
            variable_types = "$id:VecticeId!"
            variables = {"id": phase}
            kw = "id:$id"
        elif parent_id:
            gql_query = "getPhaseByName"
            variable_types = "$name:String!,$parentId:VecticeId!"
            variables = {"name": phase, "parentId": parent_id}
            kw = "name:$name,parentId:$parentId"
        else:
            raise ValueError("Missing parameters: string and parent id required.")
        query = GqlApi.build_query(
            gql_query=gql_query,
            variable_types=variable_types,
            returns=_RETURNS_FULL if full else _RETURNS,
            keyword_arguments=kw,
            query=True,
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            phase_output: PhaseOutput = Parser().parse_item(response[gql_query])
            return phase_output
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "phase", phase)

    def list_step_definitions(self, parent_id: str) -> list[StepOutput]:
        alias_filter = {"parentId": parent_id, "search": ""}
        returns = f"""items{{
                                    {_STEP_DEFINITION_RETURNS}
                        }}"""
        variable_types = "$filters:BaseDocumentationListFiltersInput!,$order:ListOrderInput"
        kw = "filters:$filters,order:$order"
        variables = {"filters": alias_filter, "order": INDEX_ORDERED}
        query = GqlApi.build_query(
            gql_query="getStepDefinitionList",
            variable_types=variable_types,
            returns=returns,
            keyword_arguments=kw,
            query=True,
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            iterations_output: list[StepOutput] = Parser().parse_list(response["getStepDefinitionList"]["items"])
            return iterations_output
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "phase", "list")



from __future__ import annotations

from typing import TYPE_CHECKING

from vectice.api._utils import read_nodejs_date
from vectice.api.json.dataset_version_representation import DatasetVersionRepresentationOutput
from vectice.models.resource.metadata.base import DatasetSourceOrigin, DatasetType

if TYPE_CHECKING:
    from datetime import datetime


class DatasetRepresentationOutput(dict):
    @property
    def created_date(self) -> datetime | None:
        return read_nodejs_date(str(self["createdDate"]))

    @property
    def updated_date(self) -> datetime | None:
        return read_nodejs_date(str(self["updatedDate"]))

    @property
    def id(self) -> str:
        return str(self["vecticeId"])

    @property
    def name(self) -> str:
        return str(self["name"])

    @property
    def type(self) -> DatasetType:
        return DatasetType[self["type"]]

    @property
    def origin(self) -> DatasetSourceOrigin:
        return DatasetSourceOrigin[self["sourceOrigin"]]

    @property
    def description(self) -> str | None:
        return str(self["description"]) if self["description"] else None

    @property
    def version(self) -> DatasetVersionRepresentationOutput:
        return DatasetVersionRepresentationOutput(**self["lastVersion"])

    @property
    def project_id(self) -> str | None:
        if "project" in self:
            return str(self["project"]["vecticeId"])
        return None



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from vectice.models.resource.base import Resource
from vectice.models.resource.metadata import DatasetSourceOrigin
from vectice.models.resource.metadata.column_metadata import DBColumn
from vectice.models.resource.metadata.dataframe_config import DataFrameType
from vectice.models.resource.metadata.db_metadata import DBMetadata, MetadataDB

if TYPE_CHECKING:
    from google.cloud.bigquery import Client as BQClient
    from google.cloud.bigquery import Table

_logger = logging.getLogger(__name__)


class BigQueryResource(Resource):
    """BigQuery resource reference wrapper.

    This resource wraps BigQuery paths that you have stored in BigQuery with optional metadata and versioning. You assign it to a step.

    ```python
    from vectice import BigQueryResource

    bq_resource = BigQueryResource(
        paths="bigquery-public-data.stackoverflow.posts_questions",
    )
    ```
    """

    _origin = DatasetSourceOrigin.BIGQUERY.value

    def __init__(
        self,
        paths: str | list[str],
        dataframes: DataFrameType | list[DataFrameType] | None = None,
        bq_client: BQClient | None = None,
    ):
        """Initialize a BigQuery resource.

        Parameters:
            paths: The paths to retrieve the datasets or the tables.
            dataframes (Optional): The dataframes allowing vectice to optionally compute more metadata about this resource such as columns stats. (Support Pandas)
            bq_client (Optional): The `google.cloud.bigquery.Client` to optionally retrieve file size, creation date and updated date (used for auto-versioning).
        """
        super().__init__(paths=paths, dataframes=dataframes)
        self.bq_client = bq_client

    def _fetch_data(self) -> dict[str, tuple[Table | None, DataFrameType | None]]:
        tables: dict[str, tuple[Table | None, DataFrameType | None]] = {}
        df_index = 0
        for path in self._paths:
            path_tables, new_df_index = self._fetch_path_data(index=df_index, path=path)
            tables.update(path_tables)
            df_index += new_df_index

        return tables

    def _fetch_path_data(
        self, index: int, path: str
    ) -> tuple[dict[str, tuple[Table | None, DataFrameType | None]], int]:
        count_dots_in_path = path.count(".")
        is_dataset = count_dots_in_path == 1
        is_table = count_dots_in_path == 2

        if is_table is True:
            return self._fetch_table_data(index=index, path=path), 1

        if is_dataset is True:
            return self._fetch_dataset_data(index=index, path=path)

        raise ValueError(f"Path '{path}' is invalid, please reference either a dataset or a table.")

    def _fetch_table_data(self, index: int, path: str) -> dict[str, tuple[Table | None, DataFrameType | None]]:
        from google.api_core.exceptions import Forbidden

        table_name = path.partition(".")[-1]

        dataframe = self._dataframes[index] if self._dataframes is not None and len(self._dataframes) > index else None
        if self.bq_client is None:
            return {table_name: (None, dataframe)}

        tb = None
        try:
            tb = self.bq_client.get_table(table=path)
        except Forbidden:
            _logger.warning(f"Failed to fetch data information for table {table_name}")
            pass

        return {table_name: (tb, dataframe)}

    def _fetch_dataset_data(
        self, index: int, path: str
    ) -> tuple[dict[str, tuple[Table | None, DataFrameType | None]], int]:
        from google.api_core.exceptions import Forbidden

        df_index = 0
        tables_dict: dict[str, tuple[Table | None, DataFrameType | None]] = {}
        dataset = path.partition(".")[-1]
        if self.bq_client is None:
            new_index = df_index + index
            dataframe = (
                self._dataframes[new_index]
                if self._dataframes is not None and len(self._dataframes) > new_index
                else None
            )
            return {dataset: (None, dataframe)}, 1
        try:
            tables_data: list[Table] = self.bq_client.list_tables(dataset=path)  # type: ignore[assignment]
            tables = sorted(tables_data, key=lambda table: table.table_id.lower())
            for table in tables:
                new_index = df_index + index
                dataframe = (
                    self._dataframes[new_index]
                    if self._dataframes is not None and len(self._dataframes) > new_index
                    else None
                )
                dataset = path.partition(".")[-1]
                table_name = f"{dataset}.{table.table_id}"
                try:
                    tables_dict[table_name] = (
                        self.bq_client.get_table(table=f"{path}.{table.table_id}"),
                        dataframe,
                    )
                except Exception:
                    _logger.warning(f"Failed to fetch data information for table {table_name}")
                    tables_dict[table_name] = None, dataframe
                    pass

                df_index += 1
            return tables_dict, df_index
        except Forbidden:
            _logger.warning(f"Failed to list tables for dataset {path}")
            return {}, 0

    def _build_metadata(self) -> DBMetadata:
        tables_metadata: dict[str, tuple[Table | None, DataFrameType | None]] = self.data
        dbs: list[MetadataDB] = []
        for table_name, [table_metadata, dataframe] in tables_metadata.items():
            columns: list[DBColumn] = []
            size = None
            rows_number = None
            updated_date = None
            created_date = None
            if table_metadata is not None:
                size = table_metadata.num_bytes
                rows_number = table_metadata.num_rows
                modified = table_metadata.modified
                updated_date = modified.isoformat() if modified is not None else None
                created = table_metadata.created
                created_date = created.isoformat() if created is not None else None
                for column in table_metadata.schema:
                    columns.append(
                        DBColumn(
                            name=column.name,
                            data_type=column.field_type,
                            is_unique=False,
                            nullable=column.is_nullable,
                            is_private_key=False,
                            is_foreign_key=False,
                        )
                    )

            dbs.append(
                MetadataDB(
                    name=table_name,
                    rows_number=rows_number,
                    size=size,
                    columns=columns,
                    created_date=created_date,
                    updated_date=updated_date,
                    dataframe=dataframe,
                )
            )

        metadata_size = None
        for db in dbs:
            if db.size is not None:
                if metadata_size is None:
                    metadata_size = 0
                metadata_size += db.size

        metadata = DBMetadata(
            size=int(metadata_size) if metadata_size else None,
            origin=self._origin,
            dbs=dbs,
        )
        return metadata



from __future__ import annotations

from vectice.models.attachment_container import AttachmentContainer
from vectice.models.errors import VecticeError
from vectice.models.git_version import CodeSource
from vectice.models.iteration import Iteration
from vectice.models.metric import Metric
from vectice.models.phase import Phase
from vectice.models.project import Project
from vectice.models.property import Property
from vectice.models.step import Step
from vectice.models.workspace import Workspace

__all__ = [
    "AttachmentContainer",
    "Metric",
    "Property",
    "Project",
    "VecticeError",
    "Workspace",
    "Phase",
    "Step",
    "Iteration",
    "CodeSource",
]



from __future__ import annotations

import inspect
from dataclasses import asdict, dataclass
from datetime import datetime, timezone


@dataclass
class MetricInput(dict):
    def __init__(self, key: str, value: float | int, timestamp: datetime | str | None = None):
        super().__init__()
        self.key: str = key
        self.value: float | int = value
        if timestamp is None:
            self.timestamp: str = datetime.now(timezone.utc).isoformat()
        else:
            self.timestamp = timestamp.isoformat() if isinstance(timestamp, datetime) else timestamp


@dataclass
class MetricOutput:
    key: str
    value: float
    timestamp: datetime
    name: str | None = None
    id: int | None = None

    @classmethod
    def from_dict(cls, metrics):
        return cls(**{k: v for k, v in metrics.items() if k in inspect.signature(cls).parameters})

    def as_dict(self):
        return asdict(self)



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from gql import gql
from gql.transport.exceptions import TransportQueryError

from vectice.api.gql_api import GqlApi, Parser
from vectice.api.json.iteration import IterationStatus
from vectice.api.json.paged_response import PagedResponse
from vectice.utils.api_utils import INDEX_ORDERED_DESC, PAGINATE_OUTPUT, get_page_input

if TYPE_CHECKING:
    from vectice.api.json.iteration import IterationInput, IterationOutput

_logger = logging.getLogger(__name__)

_RETURNS_LIST = """vecticeId
            index
            status
            owner {
                id
                name
            }
            starred
            __typename
            """

_RETURNS_PAGE = PAGINATE_OUTPUT.format(_RETURNS_LIST)

_BASE_RETURNS = """vecticeId
            index
            status
            starred
            __typename
            """
_BASE_PARENT = """
            phase {
                    vecticeId
                    name
                    status
                    index
                    __typename
              }
"""

_RETURNS = f"""
    {_BASE_RETURNS}
    {_BASE_PARENT}
"""

_PARENT_FULL = """
            phase {
                    vecticeId
                    name
                    status
                    index
                    __typename
                    parent {
                        vecticeId
                        name
                        description
                        workspace {
                            vecticeId
                            name
                            description
                            __typename
                        }
                        __typename
                    }
              }
"""

_RETURNS_FULL = f"""
    {_BASE_RETURNS}
    {_PARENT_FULL}
"""


class IterationApi(GqlApi):
    def list_iterations(
        self, parent_id: str, only_mine: bool = False, statuses: list[IterationStatus] | None = None
    ) -> PagedResponse[IterationOutput]:
        gql_query = "getIterationList"
        filters = {
            "phaseId": parent_id,
            "onlyMine": only_mine,
            "status": list(map(lambda status: status.value, statuses))
            if statuses is not None and len(statuses) > 0
            else None,
        }

        variable_types = "$filters:IterationFiltersInput!,$order:ListOrderInput,$page:PageInput"
        kw = "filters:$filters,order:$order,page:$page"
        variables = {
            "filters": filters,
            "order": INDEX_ORDERED_DESC,
            "page": get_page_input(),
        }
        query = GqlApi.build_query(
            gql_query=gql_query,
            variable_types=variable_types,
            returns=_RETURNS_PAGE,
            keyword_arguments=kw,
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            iterations_output: PagedResponse[IterationOutput] = Parser().parse_paged_response(response[gql_query])
            return iterations_output
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "iteration", "list")

    def create_iteration(self, phase_id: str) -> IterationOutput:
        gql_query = "createIteration"
        variable_types = "$phaseId:VecticeId!"
        variables = {"phaseId": phase_id}
        kw = "phaseId:$phaseId"
        query = GqlApi.build_query(
            gql_query=gql_query, variable_types=variable_types, returns=_RETURNS, keyword_arguments=kw, query=False
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            iteration_output: IterationOutput = Parser().parse_item(response[gql_query])
            return iteration_output
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "iteration", phase_id)

    def get_last_iteration(self, phase_id: str) -> IterationOutput:
        gql_query = "getLastUpdatedIterationOrCreateOne"
        variable_types = "$phaseId:VecticeId!"
        variables = {"phaseId": phase_id}
        kw = "phaseId:$phaseId"
        query = GqlApi.build_query(
            gql_query=gql_query,
            variable_types=variable_types,
            returns=_RETURNS,
            keyword_arguments=kw,
            query=True,
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            return Parser().parse_item(response[gql_query])
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "iteration", phase_id)

    def get_iteration_by_id(self, iteration_id: str, full: bool = False) -> IterationOutput:
        gql_query = "getIterationById"
        variable_types = "$id:VecticeId!"
        variables = {"id": iteration_id}
        kw = "id:$id"
        query = GqlApi.build_query(
            gql_query=gql_query,
            variable_types=variable_types,
            returns=_RETURNS_FULL if full else _RETURNS,
            keyword_arguments=kw,
            query=True,
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            iteration_output: IterationOutput = Parser().parse_item(response[gql_query])
            return iteration_output
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "iteration", iteration_id)

    def get_iteration_by_index(self, phase_id: str, index: int) -> IterationOutput:
        gql_query = "getIterationByIndex"
        variable_types = "$index:Float!,$phaseId:VecticeId!"
        variables = {"index": index, "phaseId": phase_id}
        kw = "index:$index,phaseId:$phaseId"
        query = GqlApi.build_query(
            gql_query=gql_query, variable_types=variable_types, returns=_RETURNS, keyword_arguments=kw, query=True
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            iteration_output: IterationOutput = Parser().parse_item(response[gql_query])
            return iteration_output
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "iteration_index", index)

    def update_iteration(self, iteration: IterationInput, iteration_id: str) -> IterationOutput:
        variable_types = "$id:VecticeId!,$data:IterationUpdateInput!"
        kw = "id:$id,data:$data"
        variables = {"id": iteration_id, "data": iteration}
        query = GqlApi.build_query(
            gql_query="updateIteration",
            variable_types=variable_types,
            returns=_RETURNS,
            keyword_arguments=kw,
            query=False,
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            iteration_output: IterationOutput = Parser().parse_item(response["updateIteration"])
            return iteration_output
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "iteration", "put")

    def delete_iteration(self, iteration_id: str) -> None:
        variable_types = "$id:VecticeId!"
        kw = "id:$id"
        variables = {"id": iteration_id}
        query = GqlApi.build_query(
            gql_query="removeIteration",
            variable_types=variable_types,
            keyword_arguments=kw,
            query=False,
        )
        query_built = gql(query)
        try:
            self.execute(query_built, variables)
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "iteration", iteration_id)



from __future__ import annotations

import inspect
import os
import warnings
from contextlib import contextmanager
from datetime import datetime
from functools import wraps
from typing import Any, Callable

from packaging.version import Version

from vectice.__version__ import __version__

CURRENT_VERSION = Version(Version(__version__).base_version)
_WARN_DEPR_REMOVAL = os.getenv("VECTICE_WARN_DEPR_REMOVAL", "0") == "1"


class DeprecationError(BaseException):
    """An exception raised when a deprecated object is used."""


def _exceeds(at: str | datetime | None) -> bool:
    if isinstance(at, str) and Version(at) <= CURRENT_VERSION:
        return True
    if isinstance(at, datetime) and at <= datetime.now():
        return True
    return False


def _deprecated_use(
    *,
    parameter: str | None,
    default: bool,
    args: tuple,
    kwargs: dict[str, Any],
    param_names: list[str],
) -> bool:
    # no parameter was specified for deprecation
    if parameter is None:
        return True
    # checking presence/absence of parameter in given arguments
    if default:
        # parameter must be present
        if parameter not in kwargs and len(args) <= param_names.index(parameter):
            return True
    else:
        # parameter must be absent
        if parameter in kwargs or len(args) > param_names.index(parameter):
            return True
    # no deprecated use detected
    return False


def deprecate(
    reason: str,
    *,
    parameter: str | None = None,
    default: bool = False,
    warn_at: str | datetime | None = None,
    fail_at: str | datetime | None = None,
    remove_at: str | datetime | None = None,
) -> Callable[[Callable], Callable]:
    """Deprecate a function. Internal-use only.

    This decorator can be used to deprecate a function or class method, or a parameter of this function/method.

    Parameters:
        reason: The reason why the function/method is deprecated. Useful information
            can be provided to the users, such as what to use instead of the deprecated function.
            Placeholders can be used in the reason string: `{name}`, `{parameter}`, `{warn_at}`, `{fail_at}` and `{remove_at}`:
            they will be replaced by their actual values, `name` being the name of the function.
        parameter: The name of the parameter to deprecate. If none,
            the whole function is deprecated.
        default: Whether to deprecate the default value of the given parameter (`True`),
            or the use of the parameter itself (`False`, default).
        warn_at: Version or date at which to start emitting a deprecation warning to the user.
            The user will see a deprecation warning when executing the function.
        fail_at: Version or date at which to start raising a deprecation error.
        remove_at: Internal-use only. Version or date at which to remove the deprecated code
            from the code base. It will start emitting a deprecation warning to the developers.
            The test suite will fail as long as the deprecated code is not removed.

    The `warn_at`, `fail_at` and `remove_at` values can be strings or instances of `datetime.datetime`.
    Strings are parsed as PEP 440 versions, see https://peps.python.org/pep-0440/.

    Although you don't have to always provide the three `warn_at`, `fail_at` and `remove_at` parameters,
    it is recommended to provide all three, to ensure the full life-cycle management of the deprecated function.

    IMPORTANT: Deprecation warnings are not enabled by default in Python,
    so users might need to enable them with `python -Walways`.
    More information on warnings control: https://docs.python.org/3/library/warnings.html.

    Examples:
        ```python
        # deprecating a function

        # --------------------------------------
        # vectice code in vectice/module.py
        from vectice.utils.deprecation import deprecate

        @deprecate(
            warn_at="2.1",
            fail_at="3.0",
            remove_at="4.0",
            reason="Function {name} is deprecated since version {warn_at}. "
            "Starting at version {fail_at}, it raises / will raise an exception. "
            "It will be removed at version {remove_at}. "
            "Please use 'other_function' instead.",
        )
        def my_deprecated_function(): ...
        # tests suite fails if version is >= 4.0


        # --------------------------------------
        # user code
        from vectice.module import my_deprecated_function

        my_deprecated_function()
        # exception raised if version is >= 3.0
        # warning emitted if version is >= 2.1
        # nothing happens if version is < 2.1
        ```

        ```python
        # deprecating a class method
        from vectice.utils.deprecation import deprecate


        class MyClass:
            @deprecate(warn_at="1", fail_at="2", remove_at="3", reason="This method is deprecated.")
            def my_deprecated_method(self): ...
        ```

        ```python
        # deprecating a module attribute
        from vectice.utils.deprecation import deprecate

        @deprecate(warn_at="1", fail_at="2", remove_at="3", reason="This attribute is deprecated.")
        def _get_my_deprecated_attribute():
            return "value'

        def __getattr__(name):
            if name == "my_deprecated_attribute":
                return _get_my_deprecated_attribute()
            raise AttributeError(f"module {__name__!r} has no attribute {name!r}")
        ```

        ```python
        # deprecating a class attribute
        from vectice.utils.deprecation import deprecate


        class MyClass:
            _my_deprecated_attribute = "..."

            @deprecate(warn_at="1", fail_at="2", remove_at="3", reason="This attribute is deprecated.")
            def my_deprecated_attribute(self):
                return self._my_deprecated_attribute
        ```

        ```python
        # deprecating a function parameter
        from vectice.utils.deprecation import deprecate


        @deprecate(
            parameter="my_deprecated_parameter",
            warn_at="1",
            fail_at="2",
            remove_at="3",
            reason="This parameter is deprecated.",
        )
        def my_function(my_deprecated_parameter=None): ...
        ```

        ```python
        # deprecating a parameter default value
        from vectice.utils.deprecation import deprecate


        @deprecate(
            parameter="my_parameter",
            default=True,
            warn_at="1",
            fail_at="2",
            remove_at="3",
            reason="This default value is deprecated.",
        )
        def my_function(my_parameter="my_deprecated_default_value"): ...
        ```

    Returns:
        This function returns the actual decorator that must be called again
            with the function to decorate as argument.
    """

    def decorator(func: Callable) -> Callable:
        formatted_reason = reason.format(
            name=func.__name__,
            warn_at=warn_at,
            fail_at=fail_at,
            remove_at=remove_at,
            parameter=parameter,
        )
        if _WARN_DEPR_REMOVAL and _exceeds(remove_at):
            at = f"v{remove_at}" if isinstance(remove_at, str) else remove_at
            param = (f"({parameter})" + (" (default value)" if default else "")) if parameter else ""
            warnings.warn(
                f"Reminder: {func.__module__}.{func.__qualname__}{param} has reached its End-Of-Life, "
                f"planned at {at} - you should remove it from the code base. "
                f"Deprecation reason: {formatted_reason}",
                DeprecationWarning,
                stacklevel=2,
            )

        # we make sure the parameter exists and is not required,
        # and we don't guard behind _WARN_DEPR_REMOVAL
        # so that we never fail to detect such an error,
        # to protect users from getting an exception themselves
        params = inspect.signature(func).parameters
        if parameter:
            if parameter not in params:
                raise DeprecationError(
                    f"Cannot deprecate parameter '{parameter}' in callable {func}: no such parameter"
                )
            if params[parameter].default is inspect._empty:
                default_value = "default value " if default else ""
                raise DeprecationError(
                    f"Cannot deprecate parameter '{parameter}' {default_value}in callable {func}: parameter is required"
                )

        @wraps(func)
        def wrapper(*args, **kwargs):
            if _deprecated_use(
                parameter=parameter,
                default=default,
                args=args,
                kwargs=kwargs,
                param_names=list(params),
            ):
                if _exceeds(fail_at):
                    raise DeprecationError(formatted_reason)
                if _exceeds(warn_at):
                    warnings.warn(formatted_reason, DeprecationWarning, stacklevel=2)
            return func(*args, **kwargs)

        return wrapper

    return decorator


@contextmanager
def ignore_deprecation_warnings():
    """Temporarily ignore deprecation warnings. Internal-use only.

    Yields:
        Nothing: this function is a context manager.
    """
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", DeprecationWarning)
        yield



from __future__ import annotations

import logging

from vectice.api.json.iteration import IterationStepArtifactInput
from vectice.api.json.step import StepType
from vectice.models.step import Step
from vectice.utils.common_utils import _check_read_only

_logger = logging.getLogger(__name__)


class StepString(Step):
    """Model a string step."""

    def __init__(self, step: Step, string: str | None = None):
        """Initialize a string step.

        Parameters:
            step: The step.
            string: The step string.
        """
        super().__init__(
            step.id,
            step._iteration,
            step.name,
            step.index,
            step.slug,
            step._description,
            step_type=StepType.StepString,
            artifacts=step.artifacts,
        )
        self._string = string

    def __repr__(self):
        return f"StepString(name={self.name!r}, id={self.id!r}, description={self._description!r}, string={self._string!r})"

    @property
    def string(self) -> str | None:
        """The step's string.

        Returns:
            The step's string.
        """
        return self._string

    @string.setter
    def string(self, value: str) -> None:
        """Set the step's string.

        Typical usage example:

        ```pycon
        >>> my_iteration = my_phase.create_iteration()
        ... my_iteration.step_exploration = "Step information"
        ```

        Parameters:
            value: The string to set.
        """
        _check_read_only(self.iteration)
        step_artifacts = [IterationStepArtifactInput(type="Comment", text=value)]
        self._client.update_iteration_step_artifact(self.id, StepType.StepString, step_artifacts)
        self._string = value



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, BinaryIO, NoReturn

from vectice.api.http_error_handlers import InvalidReferenceError, VecticeException
from vectice.api.json import AttachmentOutput, ModelVersionOutput, PagedResponse
from vectice.api.json.dataset_version import DatasetVersionOutput
from vectice.api.rest_api import HttpError, RestApi
from vectice.models.representation.dataset_version_representation import DatasetVersionRepresentation
from vectice.models.representation.model_version_representation import ModelVersionRepresentation

if TYPE_CHECKING:
    from io import BytesIO

    from requests import Response


MODEL_VERSION = "model version"
DATASET_VERSION = "dataset version"


_logger = logging.getLogger(__name__)


class AttachmentApi(RestApi):
    def _generate_model_url_and_id(self, model_version: ModelVersionOutput) -> tuple[str, str | None]:
        try:
            model_name = model_version.model.name
            version_name = model_version.name
            url = self._build_url(model_version.model.project_id, model_version.id, "modelversion")
            model_repr = f"Model(name='{model_name}', version='{version_name}')"
            return url, model_repr
        except HttpError as e:
            self._handle_http_error(e, model_version)

    def _generate_dataset_url_and_id(self, dataset_version: DatasetVersionOutput) -> tuple[str, str | None]:
        try:
            dataset_name = dataset_version.dataset.name
            version_name = dataset_version.name
            url = self._build_url(dataset_version.dataset.project_id, dataset_version.id, "datasetversion")
            dataset_repr = f"Dataset(name='{dataset_name}', version='{version_name}')"
            return url, dataset_repr
        except HttpError as e:
            self._handle_http_error(e, dataset_version)

    def _generate_version_url_and_id(
        self, version: ModelVersionOutput | DatasetVersionOutput
    ) -> tuple[str, str | None]:
        if isinstance(version, ModelVersionOutput):
            return self._generate_model_url_and_id(version)

        return self._generate_dataset_url_and_id(version)

    @staticmethod
    def _build_url(project_id: str | int, version: str, file_type: str) -> str:
        return f"/metadata/project/{project_id}/entityfiles/{file_type}/{version}"

    def post_attachment(
        self, files: list[tuple[str, tuple[str, BinaryIO]]], version: ModelVersionOutput | DatasetVersionOutput
    ) -> list[dict]:
        entity_files = []
        url, asset_repr = self._generate_version_url_and_id(version)
        if len(files) == 1:
            filename = files[0][1][0]
            try:
                response = self._post_attachments(url, files)
                if response:
                    entity_files.append(response.json())
                _logger.debug(f"Attachment with name: {filename} successfully attached to {asset_repr}.")
            except HttpError as e:
                self._handle_http_error(e, version, filename)
        elif len(files) > 1:
            for file in files:
                try:
                    response = self._post_attachments(url, [file])
                    if response:
                        entity_files.append(response.json())
                except HttpError as e:
                    self._handle_http_error(e, version, file[1][0])

                filenames = ", ".join([f[1][0] for f in files])
                _logger.debug(f"Attachments with names: {filenames} successfully attached to {asset_repr}.")
        return entity_files

    def post_model_predictor(self, model_type: str, model_content: BytesIO, model_version: ModelVersionOutput) -> None:
        url, model_repr = self._generate_model_url_and_id(model_version)
        url += f"?modelFramework={model_type}"
        attachment = ("file", ("model_pickle", model_content))
        self._post_attachments(url, [attachment])
        _logger.info(f"Model {model_type} successfully attached to {model_repr}.")

    def _identify_object(
        self, code_version_id: int | None = None, phase_id: str | None = None
    ) -> tuple[str, str | int]:
        if phase_id:
            return "phase", phase_id
        elif code_version_id:
            return "codeversion", code_version_id
        else:
            raise ValueError("No ID was provided for create attachment.")

    def create_attachments(
        self,
        files: list[tuple[str, tuple[str, Any]]],
        project_id: str,
        code_version_id: int | None = None,
        phase_id: str | None = None,
        step_id: int | None = None,
    ) -> list[dict]:
        parent_object, object_id = self._identify_object(code_version_id, phase_id)
        entity_files = []
        query_param = f"?stepId={step_id}" if step_id is not None else ""
        url = f"/metadata/project/{project_id}/entityfiles/{parent_object}/{object_id}{query_param}"
        try:
            for file in files:
                response = self._post_attachments(url, [file])
                if response:
                    entity_files.append(response.json())
            _logger.debug(
                f"Attachments with names: {[f[1][0] for f in files]} successfully attached to {parent_object} {object_id}."
            )
            return entity_files
        except HttpError as e:
            self._httpErrorHandler.handle_get_http_error(e, parent_object, object_id)

    def get_code_version_attachment(self, code_version_id: int, project_id: str, file_id: int) -> Response:
        try:
            return self._get_attachment(
                f"/metadata/project/{project_id}/entityfiles/codeversion/{code_version_id}/{file_id}"
            )
        except HttpError as e:
            self._handle_code_version_error(e, code_version_id)

    def list_attachments(self, version: ModelVersionOutput | DatasetVersionOutput) -> PagedResponse[AttachmentOutput]:
        try:
            url, _ = self._generate_version_url_and_id(version)
            if url is None:
                raise InvalidReferenceError(MODEL_VERSION, version.id)
            attachments = self._list_attachments(url)
        except HttpError as e:
            self._handle_http_error(e, version)
        return PagedResponse(
            item_cls=AttachmentOutput,
            total=len(attachments),
            page={},
            items=attachments,
        )

    def list_version_representation_attachments(
        self, project_id: str, version: ModelVersionRepresentation | DatasetVersionRepresentation
    ) -> PagedResponse[AttachmentOutput]:
        try:
            url_type = "modelversion" if isinstance(version, ModelVersionRepresentation) else "datasetversion"
            url = self._build_url(project_id, version.id, url_type)
            attachments = self._list_attachments(url)
        except HttpError as e:
            self._handle_http_error(e, version)

        return PagedResponse(
            item_cls=AttachmentOutput,
            total=len(attachments),
            page={},
            items=attachments,
        )

    def _handle_http_error(
        self,
        error: HttpError,
        version: ModelVersionOutput | DatasetVersionOutput | ModelVersionRepresentation | DatasetVersionRepresentation,
        file: str | None = None,
    ) -> NoReturn:
        ref_type = (
            MODEL_VERSION
            if isinstance(version, ModelVersionOutput) or isinstance(version, ModelVersionRepresentation)
            else DATASET_VERSION
        )
        if error.code == 413 and file is not None:
            raise VecticeException(f"{ref_type}: {file} exceeds the maximum size of 100MB.")
        self._httpErrorHandler.handle_get_http_error(error, ref_type, version.id)

    def _handle_code_version_error(self, error: HttpError, code_version_id: int) -> NoReturn:
        self._httpErrorHandler.handle_get_http_error(error, "code version", code_version_id)



from __future__ import annotations

import json
from json import JSONDecodeError


class Configuration:
    def __init__(self, config_file_path: str):
        with open(config_file_path, "r") as config_file:
            try:
                self.config_object = json.load(config_file)
            except JSONDecodeError as error:
                raise SyntaxError("Can not read JSON config file. Check that the file structure is valid.") from error

    def __getitem__(self, key) -> str:
        information: str = self.config_object.get(key)
        if not information:
            raise KeyError(f"Error while reading the configuration, {key} is empty.")
        return information



from __future__ import annotations

from typing import TYPE_CHECKING

from vectice.api._utils import read_nodejs_date

if TYPE_CHECKING:
    from datetime import datetime


class DatasetVersionRepresentationOutput(dict):
    @property
    def created_date(self) -> datetime | None:
        return read_nodejs_date(str(self["createdDate"]))

    @property
    def updated_date(self) -> datetime | None:
        return read_nodejs_date(str(self["updatedDate"]))

    @property
    def id(self) -> str:
        return str(self["vecticeId"])

    @property
    def name(self) -> str:
        return str(self["name"])

    @property
    def description(self) -> str | None:
        return str(self["description"]) if self["description"] else None

    @property
    def project_id(self) -> str | None:
        if "dataSet" in self:
            return str(self["dataSet"]["project"]["vecticeId"])
        return None



from __future__ import annotations

from dataclasses import InitVar, dataclass, field
from typing import Generic, Sequence, Type, TypeVar

from vectice.api.json.page import Page

ItemType = TypeVar("ItemType")


@dataclass
class PagedResponse(Generic[ItemType]):
    """Generic structure describing a result of a paged request.

    The structure contains page information and the list of items for this page.
    """

    total: int
    """Total number of available pages."""
    list: list[ItemType] = field(init=False)
    """Current list of elements for this page."""
    current_page: Page = field(init=False)
    """Information on the current page."""
    page: InitVar[dict]
    item_cls: InitVar[Type[ItemType]]
    items: InitVar[Sequence[dict]]

    def __post_init__(self, page: dict, cls: Type[ItemType], items: Sequence[dict]):
        self.current_page = Page(**page)
        self.list = []
        if items is not None:
            for item in items:
                type_item = cls(**item)
                self.list.append(type_item)



from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime, timezone


@dataclass
class Metric:
    """Define a metric for a model or a dataset.

    Parameters:
        key: The key to identify the metric.
        value: The value of the metric.
        timestamp: The timestamp of the metric. Corresponds to the
            metric creation time if not explicitly passed.
        name: The name of the metric.
    """

    key: str
    value: int | float
    timestamp: datetime | str = datetime.now(timezone.utc).isoformat()
    name: str | None = None

    def __post_init__(self):
        if not (isinstance(self.value, int) or isinstance(self.value, float)):
            raise TypeError(
                f"Metric '{self.key}' value must have type float or integer, not {type(self.value).__name__}."
            )

    def __repr__(self) -> str:
        return f"Metric(key='{self.key}', value={self.value})"



from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING

from vectice.api.json.step import PhaseOutput

if TYPE_CHECKING:
    from vectice.api.json.step import StepInput, StepOutput


class IterationStatus(Enum):
    NotStarted = "NotStarted"
    InProgress = "InProgress"
    InReview = "InReview"
    Abandoned = "Abandoned"
    Completed = "Completed"


class IterationStepArtifactType(Enum):
    ModelVersion = "ModelVersion"
    DataSetVersion = "DataSetVersion"
    EntityFile = "EntityFile"
    JobRun = "JobRun"
    Comment = "Comment"


class IterationStepArtifact(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def text(self) -> int | float | str | None:
        if self.get("text"):
            text = self["text"]
            if isinstance(text, float):
                return float(text)
            elif isinstance(text, int):
                return int(text)
            return str(text)
        return None

    @property
    def dataset_version_id(self) -> str | None:
        if self.get("datasetVersion"):
            return str(self["datasetVersion"]["vecticeId"])
        elif self.get("datasetVersionId"):
            return str(self["datasetVersionId"])
        else:
            return None

    @property
    def model_version_id(self) -> str | None:
        if self.get("modelVersion"):
            return str(self["modelVersion"]["vecticeId"])
        elif self.get("modelVersionId"):
            return str(self["modelVersionId"])
        else:
            return None

    @property
    def entity_file_id(self) -> int | None:
        if self.get("entityFileId"):
            return int(self["entityFileId"])
        else:
            return None

    @property
    def type(self) -> IterationStepArtifactType:
        return IterationStepArtifactType(self["type"])


class IterationStepArtifactInput(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def id(self) -> int | str | None:
        if self.get("id"):
            return int(self["id"]) if isinstance(self["id"], int) else str(self["id"])
        return None

    @property
    def text(self) -> int | float | str | None:
        if self.get("text"):
            text = self["text"]
            if isinstance(text, float):
                return float(text)
            elif isinstance(text, int):
                return int(text)
            return str(text)
        return None

    @property
    def type(self) -> str:
        return str(self["type"])

    @property
    def dataset_version_id(self) -> str | None:
        if self.get("datasetVersionId"):
            return str(self["datasetVersionId"])
        else:
            return None

    @property
    def model_version_id(self) -> str | None:
        if self.get("modelVersionId"):
            return str(self["modelVersionId"])
        else:
            return None

    @property
    def entity_file_id(self) -> int | None:
        if self.get("entityFileId"):
            return int(self["entityFileId"])
        else:
            return None


class IterationInput(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def steps(self) -> list[StepInput]:
        from vectice.api.json.step import StepInput

        steps_json = self["steps"]
        return [StepInput(step) for step in steps_json]


class IterationOutput(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._phase: PhaseOutput | None = None
        if "phase" in self:
            self._phase: PhaseOutput = PhaseOutput(**self["phase"])

    @property
    def id(self) -> str:
        return str(self["vecticeId"])

    @property
    def index(self) -> int:
        return int(self["index"])

    @property
    def steps(self) -> list[StepOutput]:
        from vectice.api.json.step import StepOutput

        steps_json = self["steps"]
        steps = [StepOutput(step) for step in steps_json]
        return steps

    @property
    def alias(self) -> str:
        return str(self["alias"])

    @property
    def status(self) -> IterationStatus:
        return IterationStatus(self["status"])

    @property
    def phase(self) -> PhaseOutput | None:
        return self._phase

    @property
    def starred(self) -> bool:
        return bool(self["starred"])

    @property
    def ownername(self) -> str:
        return str(self["owner"]["name"])



from vectice.api.json.model_representation import ModelRepresentationOutput
from vectice.models.representation.model_version_representation import ModelVersionRepresentation


class ModelRepresentation:
    def __init__(self, output: ModelRepresentationOutput):
        self.id = output.id
        self.project_id = output.project_id
        self.name = output.name
        self.type = output.type
        self.description = output.description
        self.version = ModelVersionRepresentation(output.version)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Generic, TypeVar

from vectice.models.resource.metadata.column_metadata import Column, Size

T = TypeVar("T")


class DataFrameWrapper(ABC, Generic[T]):
    """Base class for dataframe wrappers.

    Use DataFrameWrapper subclasses to assign wrap DataFrames to generate statistics  The
    Vectice library supports a handful of common cases.  Additional
    cases are generally easy to supply by deriving from this base
    class.  In particular, subclasses must override this class'
    abstact methods (`capture_columns()`, `get_size()`).
    """

    def __init__(self, dataframe: T):
        self.dataframe = dataframe

    @abstractmethod
    def capture_columns(self, minimum_rows_for_statistics: int) -> list[Column]:
        pass

    @abstractmethod
    def get_size(self) -> Size:
        pass



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from gql import gql
from gql.transport.exceptions import TransportQueryError

from vectice.api.gql_api import GqlApi, Parser
from vectice.api.json.dataset_representation import DatasetRepresentationOutput
from vectice.api.json.dataset_version_representation import DatasetVersionRepresentationOutput

if TYPE_CHECKING:
    from vectice.api.json.dataset_register import DatasetRegisterInput, DatasetRegisterOutput


_logger = logging.getLogger(__name__)

# TODO JobRun for lineages
_RETURNS = """
            datasetVersion {
                          vecticeId
                          name
                          dataSet {
                            name
                            projectId
                          }
            }
            useExistingDataset
            useExistingVersion
            __typename
            """

_RETURNS_DATASET = """
            vecticeId
            createdDate
            updatedDate
            name
            description
            type
            sourceOrigin
            lastVersion {
                vecticeId
                createdDate
                updatedDate
                name
                description
                __typename
            }
            project {
                vecticeId
            }
            __typename
"""

_RETURNS_DATASET_VERSION = """
            vecticeId
            createdDate
            updatedDate
            name
            description
            dataSet {
                project {
                    vecticeId
                }
            }
            __typename
"""


class GqlDatasetApi(GqlApi):
    def get_dataset(self, id: str) -> DatasetRepresentationOutput:
        variable_types = "$datasetId:VecticeId!"
        kw = "datasetId:$datasetId"
        variables = {"datasetId": id}
        query = GqlApi.build_query(
            gql_query="getDataset",
            variable_types=variable_types,
            returns=_RETURNS_DATASET,
            keyword_arguments=kw,
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            dataset_output: DatasetRepresentationOutput = Parser().parse_item(response["getDataset"])
            return dataset_output
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "dataset", id)

    def get_dataset_version(self, id: str) -> DatasetVersionRepresentationOutput:
        variable_types = "$datasetVersionId:VecticeId!"
        kw = "datasetVersionId:$datasetVersionId"
        variables = {"datasetVersionId": id}
        query = GqlApi.build_query(
            gql_query="getDatasetVersion",
            variable_types=variable_types,
            returns=_RETURNS_DATASET_VERSION,
            keyword_arguments=kw,
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            dataset_output: DatasetVersionRepresentationOutput = Parser().parse_item(response["getDatasetVersion"])
            return dataset_output
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "dataset_version", id)

    def register_dataset(
        self,
        data: DatasetRegisterInput,
        project_id: str | None = None,
        phase_id: str | None = None,
        iteration_id: str | None = None,
    ) -> DatasetRegisterOutput:
        if phase_id and project_id and not iteration_id:
            variable_types = "$projectId:VecticeId!,$phaseId:VecticeId!,$data:DatasetRegisterInput!"
            kw = "projectId:$projectId,phaseId:$phaseId,data:$data"
            variables = {"projectId": project_id, "phaseId": phase_id, "data": data}
        elif project_id and phase_id and iteration_id:
            variable_types = (
                "$projectId:VecticeId!,$phaseId:VecticeId!,$iterationId:VecticeId!,$data:DatasetRegisterInput!"
            )
            kw = "projectId:$projectId,phaseId:$phaseId,iterationId:$iterationId,data:$data"
            variables = {"projectId": project_id, "phaseId": phase_id, "iterationId": iteration_id, "data": data}
        elif project_id:
            variable_types = "$projectId:VecticeId!,$data:DatasetRegisterInput!"
            kw = "projectId:$projectId,data:$data"
            variables = {"projectId": project_id, "data": data}
        else:
            raise RuntimeError("The provided parent child ids do not match.")

        query = GqlApi.build_query(
            gql_query="registerDataset",
            variable_types=variable_types,
            returns=_RETURNS,
            keyword_arguments=kw,
            query=False,
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            dataset_output: DatasetRegisterOutput = Parser().parse_item(response["registerDataset"])
            return dataset_output
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "dataset", "register dataset")



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from gql import gql
from gql.transport.exceptions import TransportQueryError

from vectice.api.gql_api import GqlApi, Parser

if TYPE_CHECKING:
    from vectice.api.json import PagedResponse

_logger = logging.getLogger(__name__)

# TODO JobRun for lineages
_RETURNS = """
            items {
                    id
                    date
                    targetType
                    targetId
                    targetName
                    __typename
                    }
            total
            page {
                afterCursor
                hasNextPage
                }
            __typename
            """


class LastAssetApi(GqlApi):
    def get_last_assets(self, target_types: list[str], page):
        variable_types = "$targetTypes:[ActivityTargetType!],$page:PageIndexInput"
        kw = "targetTypes:$targetTypes,page:$page"
        variables = {"targetTypes": target_types, "page": page}

        query = GqlApi.build_query(
            gql_query="getLastAssets",
            variable_types=variable_types,
            returns=_RETURNS,
            keyword_arguments=kw,
            query=True,
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            assets_output: PagedResponse = Parser().parse_paged_response(response["getLastAssets"])
            return assets_output
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "asset", "getLastAssets")



from vectice.api.json.workspace import WorkspaceOutput


class UserOutput(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class UserAndDefaultWorkspaceOutput(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def default_workspace(self) -> WorkspaceOutput:
        return WorkspaceOutput(**self["defaultWorkspace:"])

    @property
    def user(self) -> UserOutput:
        return UserOutput(**self["user"])



from __future__ import annotations

from datetime import datetime

from pyspark.sql import DataFrame as SparkDF
from pyspark.sql import SparkSession
from pyspark.sql.functions import avg, col, expr, lit, max, min

from vectice.models.resource.metadata.column_metadata import (
    BooleanStat,
    ColumnCategoryType,
    DateStat,
    MostCommon,
    NumericalStat,
    Quantiles,
    Size,
    TextStat,
)
from vectice.models.resource.metadata.df_wrapper_default import DataFrameDefaultWrapper


class SparkDFWrapper(DataFrameDefaultWrapper[SparkDF]):
    def __init__(self, dataframe: SparkDF):
        super().__init__(dataframe)
        self.spark = SparkSession.builder.getOrCreate()

    def get_size(self) -> Size:
        self.rows: int = self.dataframe.count()
        self.columns_numbers: int = len(self.dataframe.columns)

        return Size(rows=self.rows, columns=self.columns_numbers)

    def __capture_column_stats__(self, column: SparkDF, minimum_rows_for_statistics: int):
        column_type = column.dtypes[0][1]

        if column_type == "boolean":
            return (
                ColumnCategoryType.BOOLEAN,
                self.__compute_boolean_column_statistics__(column)
                if self.rows >= minimum_rows_for_statistics
                else None,
            )
        elif column_type in ["tinyint", "smallint", "int", "bigint", "float", "double", "decimal"]:
            return (
                ColumnCategoryType.NUMERICAL,
                self.__compute_numeric_column_statistics__(column)
                if self.rows >= minimum_rows_for_statistics
                else None,
            )
        elif column_type == "date" or column_type == "timestamp":
            return (
                ColumnCategoryType.DATE,
                self.__compute_date_column_statistics__(column) if self.rows >= minimum_rows_for_statistics else None,
            )
        elif column_type in ["string", "char", "varchar"]:
            return (
                ColumnCategoryType.TEXT,
                self.__compute_string_column_statistics__(column) if self.rows >= minimum_rows_for_statistics else None,
            )

        return None, None

    def __compute_boolean_column_statistics__(self, column: SparkDF):
        true_count = column.filter(column[0] == lit(True)).count()
        false_count = column.filter(column[0] == lit(False)).count()
        missing_count = column.filter(column[0].isNull()).count()

        total_count = self.dataframe.count()
        true_percentage = true_count / total_count if total_count > 0 else 0.0
        false_percentage = false_count / total_count if total_count > 0 else 0.0
        missing_percentage = missing_count / total_count if total_count > 0 else 0.0

        return BooleanStat(
            true=float(true_percentage), false=float(false_percentage), missing=float(missing_percentage)
        )

    def __compute_numeric_column_statistics__(self, column: SparkDF):
        col_name = column.columns[0]

        count = column.count()
        missing_count = column.filter(column[0].isNull()).count()
        missing_percentage = missing_count / count if count > 0 else 0.0

        statistics = column.select(col_name).summary("mean", "stddev", "min", "25%", "50%", "75%", "max")
        statistics_list = list(map(lambda row: row.asDict(), statistics.collect()))

        summary_dict = {data["summary"]: float(data[col_name]) for data in statistics_list}

        mean = summary_dict["mean"]
        std_deviation = summary_dict["stddev"]
        min_value = summary_dict["min"]
        q25 = summary_dict["25%"]
        q50 = summary_dict["50%"]
        q75 = summary_dict["75%"]
        max_value = summary_dict["max"]

        return NumericalStat(
            mean=float(mean),
            std_deviation=float(std_deviation),
            quantiles=Quantiles(
                q_min=float(min_value), q25=float(q25), q50=float(q50), q75=float(q75), q_max=float(max_value)
            ),
            missing=float(missing_percentage),
        )

    def __compute_string_column_statistics__(self, column: SparkDF):
        count = column.count()
        unique_count = column.distinct().count()

        value_counts = column.groupBy(column[0]).count().orderBy(col("count").desc()).limit(3).collect()
        most_commons = [MostCommon(str(row[0]), float(row[1] / count)) for row in value_counts]

        missing_count = column.filter(column[0].isNull()).count()
        missing_percentage = missing_count / count if count > 0 else 0.0

        return TextStat(unique=float(unique_count), missing=float(missing_percentage), most_commons=most_commons)

    def __compute_date_column_statistics__(self, column: SparkDF):
        count = column.count()
        min_date = column.select(min(column[0])).collect()[0][0].isoformat()
        max_date = column.select(max(column[0])).collect()[0][0].isoformat()

        col_name = column.columns[0]

        df = column.withColumn(col_name, column[col_name].cast("timestamp"))
        mean = datetime.isoformat(df.agg(avg(col_name).cast("timestamp").alias("avg_created_datetime")).collect()[0][0])
        df = column.withColumn(col_name, expr("UNIX_TIMESTAMP({}, 'yyyy-MM-dd')".format(col_name)))
        median = df.approxQuantile(col_name, [0.5], 0.0)[0]
        median_iso = datetime.strptime(
            self.spark.sql(f"SELECT from_unixtime({median!s}, 'yyyy-MM-dd')").collect()[0][0], "%Y-%m-%d"
        ).isoformat()

        return DateStat(
            missing=float(column.filter(column[0].isNull()).count() / count) if count > 0 else 0.0,
            minimum=str(min_date),
            mean=str(mean),
            median=str(median_iso),
            maximum=str(max_date),
        )



from __future__ import annotations

import logging
from textwrap import dedent
from typing import TYPE_CHECKING

from rich.table import Table

from vectice.models.project import Project
from vectice.utils.common_utils import _temp_print
from vectice.utils.logging_utils import format_description

if TYPE_CHECKING:
    from vectice import Connection
    from vectice.api import Client


_logger = logging.getLogger(__name__)


class Workspace:
    """Represent a Vectice Workspace.

    Workspaces are containers used to organize projects, assets, and
    users.

    Vectice users have access to a personal workspace and other
    workspaces so they can learn and collaborate with other users. An
    organization will have many workspaces, each with an Admin and
    Members with different privileges.

    Note that only an Org Admin can create a new workspace in the
    organization.

    You can get a workspace from your [`Connection`][vectice.Connection]
    object by calling `workspace()`:

    ```python
    import vectice

    connection = vectice.connect(...)
    my_workspace = connection.workspace("Iris workspace")
    ```

    Or you can get it directly when connecting to Vectice:

    ```python
    import vectice

    my_workspace = vectice.connect(..., workspace="Iris workspace")
    ```

    See [`Connection.connect`][vectice.Connection.connect] to learn
    how to connect to Vectice.
    """

    def __init__(self, id: str, name: str, description: str | None = None):
        """Initialize a workspace.

        Vectice users shouldn't need to instantiate Workspaces manually,
        but here are the workspace parameters.

        Parameters:
            id: The workspace identifier.
            name: The name of the workspace.
            description: The description of the workspace.
        """
        self._id = id
        self._name = name
        self._description = description
        self._client: Client
        self._connection: Connection

    def __post_init__(self, client: Client, connection: Connection):
        self._client = client
        self._connection = connection

    def __eq__(self, other: object):
        if not isinstance(other, Workspace):
            return NotImplemented
        return self.id == other.id

    def __repr__(self):
        description = self._description if self._description else "None"
        return f"Workspace(name={self.name!r}, id={self._id}, description={description!r})"

    @property
    def id(self) -> str:
        """The workspace's id.

        Returns:
            The workspace's id.
        """
        return self._id

    @property
    def name(self) -> str:
        """The workspace's name.

        Returns:
            The workspace's name.
        """
        return self._name

    @property
    def description(self) -> str | None:
        """The workspace's description.

        Returns:
            The workspace's description.
        """
        return self._description

    @property
    def properties(self) -> dict:
        """The workspace's name and id.

        Returns:
            A dictionary containing the `name` and `id` items.
        """
        return {"name": self.name, "id": self.id}

    def project(self, project: str) -> Project:
        """Get a project.

        Parameters:
            project: The project name or id.

        Returns:
            The project.
        """
        item = self._client.get_project(project, self.id)
        logging_output = dedent(
            f"""
                Project {item.name!r} successfully retrieved."

                For quick access to the Project in the Vectice web app, visit:
                {self._client.auth._API_BASE_URL}/browse/project/{item.id}"""
        ).lstrip()
        _logger.info(logging_output)
        project_object = Project(item.id, self, item.name, item.description)
        return project_object

    def list_projects(self) -> None:
        """Prints a list of projects belonging to the workspace in a tabular format, limited to the first 10 items. A link is provided to view the remaining projects.

        Returns:
            None
        """
        project_outputs = self._client.list_projects(self.id)
        rich_table = Table(expand=True, show_edge=False)

        rich_table.add_column("Project id", justify="left", no_wrap=True, min_width=4, max_width=10)
        rich_table.add_column("Name", justify="left", no_wrap=True, max_width=15)
        rich_table.add_column("Description", justify="left", no_wrap=True, max_width=50)

        for project in project_outputs.list:
            rich_table.add_row(project.id, project.name, format_description(project.description))

        description = dedent(
            f"""
        There are {project_outputs.total} projects in the workspace {self.name!r} and a maximum of 10 projects are displayed in the table below:
        """
        ).lstrip()
        tips = dedent(
            """
        To access a specific project, use \033[1mworkspace\033[0m.project(Project ID)"""
        ).lstrip()
        link = dedent(
            f"""
            For quick access to the list of projects in the Vectice web app, visit:
            {self._client.auth._API_BASE_URL}/browse/workspace/{self.id}"""
        ).lstrip()

        _temp_print(description)
        _temp_print(table=rich_table)
        _temp_print(tips)
        _temp_print(link)

    @property
    def projects(self) -> list[Project]:
        """List projects.

        Returns:
            A list of projects in this workspace.
        """
        response = self._client.list_projects(self.id)
        return [Project(item.id, self, item.name, item.description) for item in response.list]

    @property
    def connection(self) -> Connection:
        """The Connection to which this workspace belongs.

        Returns:
            The Connection to which this workspace belongs.
        """
        return self._connection



from __future__ import annotations

from enum import Enum


class PublicConfigOutput(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def versions(self) -> list["VersionOutput"]:
        return [VersionOutput(**version) for version in self["versions"]]


class VersionOutput(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def version(self) -> str:
        return str(self["version"])

    @property
    def artifact_name(self) -> "ArtifactName":
        return ArtifactName(self["artifactName"])


class ArtifactName(Enum):
    GLOBAL = "GLOBAL"
    BACKEND = "BACKEND"



from __future__ import annotations

import logging
import re
from datetime import datetime
from typing import TYPE_CHECKING, BinaryIO

from gql import Client as GQLClient
from gql.transport.requests import RequestsHTTPTransport

from vectice.__version__ import __version__
from vectice.api._auth import Auth
from vectice.api.attachment import AttachmentApi
from vectice.api.compatibility import CompatibilityApi
from vectice.api.gql_code import GqlCodeApi
from vectice.api.gql_code_version import GqlCodeVersionApi
from vectice.api.gql_dataset import GqlDatasetApi
from vectice.api.gql_entity_file import GqlEntityFileApi
from vectice.api.gql_feature_flag import GqlFeatureFlagApi
from vectice.api.gql_model import GqlModelApi
from vectice.api.gql_organization import GqlOrganizationApi
from vectice.api.gql_user_workspace_api import UserAndDefaultWorkspaceApi
from vectice.api.http_error_handlers import MissingReferenceError, VecticeException
from vectice.api.iteration import IterationApi
from vectice.api.json import (
    ArtifactName,
    CodeInput,
    CodeVersionCreateBody,
    ModelRegisterInput,
    ModelRegisterOutput,
    ModelType,
    ModelVersionOutput,
    ModelVersionStatus,
    PagedResponse,
    PropertyInput,
    StepOutput,
)
from vectice.api.json.dataset_register import DatasetRegisterInput, DatasetRegisterOutput
from vectice.api.json.dataset_representation import DatasetRepresentationOutput
from vectice.api.json.dataset_version import DatasetVersionOutput
from vectice.api.json.dataset_version_representation import DatasetVersionRepresentationOutput
from vectice.api.json.iteration import IterationStatus
from vectice.api.json.metric import MetricInput
from vectice.api.json.model_representation import ModelRepresentationOutput
from vectice.api.json.model_version_representation import ModelVersionRepresentationOutput
from vectice.api.json.organization_config import OrgConfigOutput
from vectice.api.last_assets import LastAssetApi
from vectice.api.phase import PhaseApi
from vectice.api.project import ProjectApi
from vectice.api.step import StepApi
from vectice.api.version import VersionApi
from vectice.api.workspace import WorkspaceApi
from vectice.models.dataset import Dataset
from vectice.models.representation.dataset_version_representation import DatasetVersionRepresentation
from vectice.models.representation.model_version_representation import ModelVersionRepresentation
from vectice.models.resource.metadata.base import MetadataSettings
from vectice.utils.vectice_ids_regex import WORKSPACE_VID_REG

if TYPE_CHECKING:
    from io import BytesIO, IOBase

    from requests import Response

    from vectice.api.json import AttachmentOutput, ProjectOutput, WorkspaceOutput
    from vectice.api.json.compatibility import CompatibilityOutput
    from vectice.api.json.iteration import IterationInput, IterationOutput, IterationStepArtifactInput
    from vectice.api.json.phase import PhaseOutput
    from vectice.api.json.step import StepType
    from vectice.models.model import Model

_logger = logging.getLogger(__name__)


DISABLED_FEATURE_FLAG_MESSAGE = (
    "This '{}' feature is not enabled. Please contact your Account Manager for Beta program access."
)


class Client:
    """Low level Vectice API client."""

    def __init__(
        self,
        token: str,
        api_endpoint: str,
    ):
        self.auth = Auth(api_endpoint=api_endpoint, api_token=token)
        transport = RequestsHTTPTransport(url=self.auth.api_base_url + "/graphql", verify=self.auth.verify_certificate)
        logging.getLogger("gql.transport.requests").setLevel("WARNING")
        self._gql_client = GQLClient(transport=transport)
        self._logger = logging.getLogger(self.__class__.__name__)
        self._org_config = self._get_org_config()

    @property
    def version_api(self) -> str:
        return __version__

    @property
    def version_backend(self) -> str:
        versions = VersionApi(self._gql_client, self.auth).get_public_config().versions
        for version in versions:
            if version.artifact_name == ArtifactName.BACKEND:
                return version.version
        raise ValueError("No version found for backend.")

    def check_compatibility(self) -> CompatibilityOutput:
        return CompatibilityApi(self.auth).check_version()

    def _get_org_config(self) -> OrgConfigOutput:
        return GqlOrganizationApi(self._gql_client, self.auth).get_organization_config()

    def list_projects(
        self,
        workspace: str,
    ) -> PagedResponse[ProjectOutput]:
        """List the projects in a workspace.

        Parameters:
            workspace: The workspace id.

        Returns:
            The workspace's projects.
        """
        return ProjectApi(self._gql_client, self.auth).list_projects(workspace)

    def get_project(self, project: str, workspace: str | None = None) -> ProjectOutput:
        """Get a project.

        Parameters:
            project: The project name or vectice id.
            workspace: The workspace name or id.

        Returns:
            The project JSON structure.
        """
        if workspace is not None and not re.search(WORKSPACE_VID_REG, workspace):
            workspace = WorkspaceApi(self._gql_client, self.auth).get_workspace(workspace).id
        return ProjectApi(self._gql_client, self.auth).get_project(project, workspace)

    def get_workspace(self, workspace: str) -> WorkspaceOutput:
        """Get a workspace.

        Parameters:
            workspace: The workspace name or id.

        Returns:
            The workspace JSON structure.
        """
        return WorkspaceApi(self._gql_client, self.auth).get_workspace(workspace)

    def list_workspaces(self) -> PagedResponse[WorkspaceOutput]:
        """List the workspaces.

        Returns:
            The workspaces.
        """
        return WorkspaceApi(self._gql_client, self.auth).list_workspaces()

    def create_code_attachments(self, files: list[tuple[str, tuple[str, str]]], code_version_id: int, project_id: str):
        """Create an attachment.

        Parameters:
            files: The paths to the files to attach.
            code_version_id: The code version id to attach files to.
            project_id: The project id associated to the code version id.

        Returns:
            The JSON structure.
        """
        return AttachmentApi(self.auth).create_attachments(files, project_id, code_version_id=code_version_id)

    def create_version_attachments(
        self, files: list[tuple[str, tuple[str, BinaryIO]]], version: ModelVersionOutput | DatasetVersionOutput
    ):
        """Create an attachment.

        Parameters:
            files: The paths to the files to attach.
            version: The version to attach files to.

        Returns:
            The JSON structure.
        """
        return AttachmentApi(self.auth).post_attachment(files, version)

    def create_phase_attachments(
        self,
        files: list[tuple[str, tuple[str, BytesIO | IOBase]]],
        phase_id: str,
        project_id: str,
        step_id: int | None = None,
    ) -> list[dict]:
        """Create an attachment.

        Parameters:
            files: The paths to the files to attach.
            phase_id: The phase id to attach files to.
            project_id: The project id to attach files to.
            step_id (Optional): The step id to add this attachment as an artifact of the step.

        Returns:
            The JSON structure.
        """
        return AttachmentApi(self.auth).create_attachments(files, project_id, phase_id=phase_id, step_id=step_id)

    def create_model_predictor(self, model_type: str, model_content: BytesIO, model_version: ModelVersionOutput):
        """Create a predictor.

        Parameters:
            model_type: The type of model to attach.
            model_content: The binary content of the model.
            model_version: The model version to attach files to.

        Returns:
            The JSON structure.
        """
        return AttachmentApi(self.auth).post_model_predictor(model_type, model_content, model_version)

    def list_version_representation_attachments(
        self, project_id: str, version: ModelVersionRepresentation | DatasetVersionRepresentation
    ) -> PagedResponse[AttachmentOutput]:
        """List the attachments of an artifact.

        Parameters:
            version: The version to list attachments from.

        Returns:
            The attachments of an artifact.
        """
        return AttachmentApi(self.auth).list_version_representation_attachments(project_id, version)

    def list_attachments(self, version: ModelVersionOutput | DatasetVersionOutput) -> PagedResponse[AttachmentOutput]:
        """List the attachments of an artifact.

        Parameters:
            version: The version to list attachments from.

        Returns:
            The attachments of an artifact.
        """
        return AttachmentApi(self.auth).list_attachments(version)

    def get_code_version_attachment(self, code_version_id: int, project_id: str, file_id: int) -> Response:
        """Get the attachment of a code version.

        Parameters:
            code_version_id: The code version id to list attachments from.
            project_id: The project id the code version belongs to.
            file_id: The file id attached to the code version.

        Returns:
            The file attached to the code version.
        """
        return AttachmentApi(self.auth).get_code_version_attachment(code_version_id, project_id, file_id)

    def list_phases(
        self,
        project: str,
    ) -> PagedResponse[PhaseOutput]:
        return PhaseApi(self._gql_client, self.auth).list_phases(project)

    def get_phase(self, phase: str, project_id: str | None = None) -> PhaseOutput:
        if project_id is None:
            raise MissingReferenceError("project")
        return PhaseApi(self._gql_client, self.auth).get_phase(phase, project_id)

    def get_full_phase(self, phase: str) -> PhaseOutput:
        return PhaseApi(self._gql_client, self.auth).get_phase(phase=phase, full=True)

    def get_step_by_name(self, step_reference: str, iteration_id: str) -> StepOutput:
        return StepApi(self._gql_client, self.auth).get_step(step_reference, iteration_id)

    def list_steps(self, iteration_id: str, populate: bool = True) -> PagedResponse[StepOutput]:
        return StepApi(self._gql_client, self.auth).list_steps(iteration_id, populate)

    def add_iteration_step_artifact(self, step_id: int, artifact: IterationStepArtifactInput) -> StepOutput:
        return StepApi(self._gql_client, self.auth).add_iteration_step_artifact(artifact, step_id)

    def update_iteration_step_artifact(
        self,
        step_id: int,
        step_type: StepType,
        artifacts: list[IterationStepArtifactInput] | None = None,
    ) -> StepOutput:
        return StepApi(self._gql_client, self.auth).update_iteration_step_artifact(step_id, step_type, artifacts)

    def list_iterations(
        self, phase: str, only_mine: bool = False, statuses: list[IterationStatus] | None = None
    ) -> PagedResponse[IterationOutput]:
        return IterationApi(self._gql_client, self.auth).list_iterations(phase, only_mine, statuses)

    def list_step_definitions(self, phase: str) -> list[StepOutput]:
        return PhaseApi(self._gql_client, self.auth).list_step_definitions(phase)

    def get_last_iteration(self, phase_id: str) -> IterationOutput:
        return IterationApi(self._gql_client, self.auth).get_last_iteration(phase_id)

    def get_iteration_by_id(self, iteration_id: str, full: bool = False) -> IterationOutput:
        return IterationApi(self._gql_client, self.auth).get_iteration_by_id(iteration_id, full)

    def get_iteration_by_index(self, phase_id: str, index: int) -> IterationOutput:
        return IterationApi(self._gql_client, self.auth).get_iteration_by_index(phase_id, index)

    def create_iteration(self, phase_id: str) -> IterationOutput:
        return IterationApi(self._gql_client, self.auth).create_iteration(phase_id)

    def update_iteration(self, iteration_id: str, iteration: IterationInput) -> IterationOutput:
        return IterationApi(self._gql_client, self.auth).update_iteration(iteration, iteration_id)

    def delete_iteration(self, iteration_id: str) -> None:
        IterationApi(self._gql_client, self.auth).delete_iteration(iteration_id)

    def register_dataset_from_source(
        self,
        dataset: Dataset,
        project_id: str | None = None,
        phase_id: str | None = None,
        iteration_id: str | None = None,
        code_version_id: int | None = None,
    ) -> DatasetRegisterOutput:
        if dataset._has_bigquery_resource and self.is_feature_flag_enabled("bigquery-dataset-source") is False:
            raise VecticeException(DISABLED_FEATURE_FLAG_MESSAGE.format("bigquery-dataset-source"))

        if dataset._has_dataframe is True and self.is_feature_flag_enabled("dataset-dataframe") is False:
            raise VecticeException(DISABLED_FEATURE_FLAG_MESSAGE.format("dataset-dataframe"))

        if dataset._has_spark_df is True and self.is_feature_flag_enabled("dataset-spark-dataframe") is False:
            raise VecticeException(DISABLED_FEATURE_FLAG_MESSAGE.format("dataset-spark-dataframe"))

        name = self.get_dataset_name(dataset)
        derived_from = self.get_derived_from(dataset)
        resources = dataset.resource if isinstance(dataset.resource, tuple) else [dataset.resource]
        metadata_asdict = []
        for resource in resources:
            if resource is not None:
                resource.metadata.set_settings(
                    MetadataSettings(
                        minimum_rows_for_statistics=self._org_config.configuration.df_statistics_row_threshold
                    )
                )
                metadata_asdict.append(resource.metadata.asdict())

        properties = (
            [vars(PropertyInput(prop.key, prop.value)) for prop in dataset.properties] if dataset.properties else None
        )

        dataset_register_input = DatasetRegisterInput(
            name=name,
            type=dataset.type.value,
            datasetSources=metadata_asdict,
            inputs=derived_from,
            codeVersionId=code_version_id,
            properties=properties,
        )
        dataset_register_output = self.register_dataset(dataset_register_input, project_id, phase_id, iteration_id)
        dataset.latest_version_id = dataset_register_output["datasetVersion"]["vecticeId"]
        return dataset_register_output

    def get_dataset(self, id: str) -> DatasetRepresentationOutput:
        return GqlDatasetApi(self._gql_client, self.auth).get_dataset(id)

    def get_dataset_version(self, id: str) -> DatasetVersionRepresentationOutput:
        return GqlDatasetApi(self._gql_client, self.auth).get_dataset_version(id)

    def get_model(self, id: str) -> ModelRepresentationOutput:
        return GqlModelApi(self._gql_client, self.auth).get_model(id)

    def get_model_version(self, id: str) -> ModelVersionRepresentationOutput:
        return GqlModelApi(self._gql_client, self.auth).get_model_version(id)

    @staticmethod
    def get_dataset_name(dataset: Dataset) -> str:
        return f"dataset {datetime.time}" if dataset.name is None else dataset.name

    @staticmethod
    def get_derived_from(obj: Dataset | Model) -> list[str]:
        return [] if obj.derived_from is None else obj.derived_from

    def register_dataset(
        self,
        dataset_register_input: DatasetRegisterInput,
        project_id: str | None = None,
        phase_id: str | None = None,
        iteration_id: str | None = None,
    ) -> DatasetRegisterOutput:
        data: DatasetRegisterOutput = GqlDatasetApi(self._gql_client, self.auth).register_dataset(
            dataset_register_input, project_id, phase_id, iteration_id
        )
        _logger.debug(
            f"Successfully registered Dataset("
            f"name='{dataset_register_input.name}', "
            f"id={data['datasetVersion']['vecticeId']}, "
            f"version='{data['datasetVersion']['name']}', "
            f"type={dataset_register_input.type})."
        )
        return data

    @staticmethod
    def _get_model_name(library: str, technique: str, name: str | None = None) -> str:
        return name if name else f"{library} {technique} model"

    def register_model(
        self,
        model: Model,
        project_id: str,
        phase_id: str | None = None,
        iteration_id: str | None = None,
        code_version_id: int | None = None,
    ) -> ModelRegisterOutput:
        """Register a model.

        Parameters:
            model: The model to register
            project_id: The project ID
            phase_id: The phase ID
            iteration_id: The iteration ID
            code_version_id: The code version ID

        Returns:
            The registered model.
        """
        metrics = [vars(MetricInput(metric.key, metric.value)) for metric in model.metrics] if model.metrics else None
        properties = (
            [vars(PropertyInput(prop.key, prop.value)) for prop in model.properties] if model.properties else None
        )
        derived_from = self.get_derived_from(model)
        model_register_input = ModelRegisterInput(
            name=model.name,
            modelType=ModelType.OTHER.value,
            status=ModelVersionStatus.EXPERIMENTATION.value,
            inputs=derived_from,
            metrics=metrics,
            properties=properties,
            algorithmName=model.technique,
            framework=model.library,
            codeVersionId=code_version_id,
        )
        return GqlModelApi(self._gql_client, self.auth).register_model(
            model_register_input, project_id, phase_id, iteration_id
        )

    def get_last_assets(self, target_types: list[str], page):
        return LastAssetApi(self._gql_client, self.auth).get_last_assets(target_types, page)

    def get_user_and_default_workspace(self):
        return UserAndDefaultWorkspaceApi(self._gql_client, self.auth).get_user_and_default_workspace()

    def create_code_gql(self, project_id: str, code: CodeInput):
        return GqlCodeApi(self._gql_client, self.auth).create_code(project_id, code)

    def create_code_version_gql(self, code_id: int, code_version: CodeVersionCreateBody):
        return GqlCodeVersionApi(self._gql_client, self.auth).create_code_version(code_id, code_version)

    def get_code(self, code: str | int, project_id: str | None = None):
        if project_id is None:
            raise MissingReferenceError("project")
        return GqlCodeApi(self._gql_client, self.auth).get_code(code, project_id)

    def get_code_version(self, code_version: str | int, code_id: int | None = None):
        if code_id is None:
            raise MissingReferenceError("code")
        return GqlCodeVersionApi(self._gql_client, self.auth).get_code_version(code_version, code_id)

    def get_entity_file_by_id(self, id: int):
        return GqlEntityFileApi(self._gql_client, self.auth).get_entity_file_by_id(id)

    def is_feature_flag_enabled(self, code: str) -> bool:
        enabled = GqlFeatureFlagApi(self._gql_client, self.auth).is_feature_flag_enabled(code)
        if enabled is False:
            _logger.info(DISABLED_FEATURE_FLAG_MESSAGE.format(code))
        return enabled



from __future__ import annotations

import logging
from io import IOBase

from PIL.Image import Image

from vectice.api.json.step import StepType
from vectice.models.step import Step

_logger = logging.getLogger(__name__)


class StepImage(Step):
    """Model a Vectice step's image.

    A StepImage stores an image.
    """

    def __init__(self, step: Step, image: str | IOBase | Image | None = None):
        """Initialize a step's image.

        Parameters:
            step: The step
            image: The step's image.
        """
        super().__init__(
            id=step.id,
            iteration=step._iteration,
            name=step.name,
            index=step.index,
            slug=step.slug,
            description=step._description,
            step_type=StepType.StepImage,
            artifacts=step.artifacts,
        )
        self._image: str | IOBase | Image | None = image

    def __repr__(self):
        return f"StepImage(name='{self.name}', id={self.id}, description='{self._description}', image={self.image})"

    @property
    def image(self) -> str | IOBase | Image | None:
        """The step's image.

        Returns:
            The step's image.
        """
        return self._image



from __future__ import annotations

from dataclasses import dataclass


@dataclass
class UserDeclaredVersion:
    name: str | None = None
    uri: str | None = None
    description: str | None = None
    isStarred: bool | None = False



from vectice.api.json.dataset_representation import DatasetRepresentationOutput
from vectice.models.representation.dataset_version_representation import DatasetVersionRepresentation


class DatasetRepresentation:
    def __init__(self, output: DatasetRepresentationOutput):
        self.id = output.id
        self.project_id = output.project_id
        self.name = output.name
        self.type = output.type
        self.origin = output.origin
        self.description = output.description
        self.version = DatasetVersionRepresentation(output.version)



from __future__ import annotations

from typing import TYPE_CHECKING

from vectice.api._utils import read_nodejs_date
from vectice.api.json.project import ProjectOutput

if TYPE_CHECKING:
    from datetime import datetime


class CodeInput(dict):
    @property
    def code_id(self) -> int:
        return int(self["codeId"])

    @property
    def name(self) -> str:
        return str(self["name"])

    @property
    def description(self) -> str:
        return str(self["description"])

    @property
    def uri(self) -> str:
        return str(self["uri"])


class CodeOutput(dict):
    _project: ProjectOutput

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if "project" in self:
            self._project = ProjectOutput(**self["project"])
        else:
            self._project = None

    @property
    def created_date(self) -> datetime | None:
        return read_nodejs_date(str(self["createdDate"]))

    @property
    def updated_date(self) -> datetime | None:
        return read_nodejs_date(str(self["updatedDate"]))

    @property
    def id(self) -> int:
        return int(self["id"])

    @property
    def name(self) -> str:
        return str(self["name"])

    @property
    def description(self) -> str | None:
        if "description" in self and self["description"] is not None:
            return str(self["description"])
        else:
            return None

    @property
    def uri(self) -> str:
        return str(self["uri"])

    @property
    def author_id(self) -> int:
        return int(self["authorId"])

    @property
    def deleted_date(self) -> datetime | None:
        return read_nodejs_date(str(self["deletedDate"]))

    @property
    def version_number(self) -> int:
        return int(self["versionNumber"])

    @property
    def project(self) -> ProjectOutput:
        return self._project

    @project.setter
    def project(self, project: ProjectOutput):
        self._project = project

    @property
    def project_id(self) -> int | None:
        return int(self["projectId"])



from __future__ import annotations


class OrganizationConfigOutput(dict):
    @property
    def df_statistics_row_threshold(self) -> int:
        return int(self["dfStatisticsRowThreshold"])


class OrgConfigOutput(dict):
    @property
    def configuration(self) -> OrganizationConfigOutput:
        return OrganizationConfigOutput(self["organization"]["configuration"])



from __future__ import annotations

import datetime
import glob
import hashlib
import logging
import os

from vectice.models.resource.base import Resource
from vectice.models.resource.metadata.base import DatasetSourceOrigin
from vectice.models.resource.metadata.dataframe_config import DataFrameType
from vectice.models.resource.metadata.files_metadata import File, FilesMetadata

_logger = logging.getLogger(__name__)


class FileResource(Resource):
    """Wrap columnar data and its metadata in a local file.

    Vectice stores metadata -- data about your dataset -- communicated
    with a resource.  Your actual dataset is not stored by Vectice.

    This resource wraps data that you have stored in a local file.

    ```python
    from vectice import FileResource, connect

    my_project = connect(...)  # (1)
    my_phase = my_project.phase(...)  # (2)
    my_iter = my_phase.create_iteration()  # (3)

    my_iter.step_my_data = FileResource(paths="my/file/path")
    ```

    1. See [connection][vectice.Connection.connect].
    1. See [phases][vectice.models.Phase].
    1. See [iterations][vectice.models.Iteration].

    Note that these three concepts are distinct, even if easily conflated:

    * Where the data is stored
    * The format at rest (in storage)
    * The format when loaded in a running Python program

    Notably, the statistics collectors provided by Vectice operate
    only on this last and only in the event that the data is loaded as
    a pandas dataframe.
    """

    _origin = DatasetSourceOrigin.LOCAL.value

    def __init__(
        self,
        paths: str | list[str],
        dataframes: DataFrameType | list[DataFrameType] | None = None,
    ):
        """Initialize a file resource.

        Parameters:
            paths: The paths of the files to wrap.
            dataframes (Optional): The dataframes allowing vectice to optionally compute more metadata about this resource such as columns stats. (Support Pandas)

        Examples:
            The following example shows how to wrap a CSV file
            called `iris.csv` in the current directory:

            ```python
            from vectice import FileResource
            iris_trainset = FileResource(paths="iris.csv")
            ```
        """
        super().__init__(paths=paths, dataframes=dataframes)
        paths_log = ", ".join(self._paths) if len(self._paths) > 1 else self._paths[0]
        _logger.info(f"File: {paths_log} wrapped successfully.")

    def _fetch_data(self) -> dict[str, bytes]:
        datas = {}
        metadata: FilesMetadata = self.metadata  # type:ignore[assignment]
        for file in metadata.files:
            with open(file.name, "rb") as opened_file:
                datas[file.name] = opened_file.read()
        return datas

    def _build_metadata(self) -> FilesMetadata:
        if self._paths is None:
            raise ValueError("Paths can not be None.")
        size = None
        files = []
        df_index = 0
        for path in self._paths:
            new_files, total_size, new_df_index = self._file_visitor_list_builder(index=df_index, path=path)
            df_index += new_df_index
            if size is None:
                size = 0
            size += total_size
            files.extend(new_files)

        metadata = FilesMetadata(size=size, origin=self._origin, files=files)
        return metadata

    def _file_visitor_list_builder(self, index: int, path: str) -> tuple[list[File], int, int]:
        files: list[File] = []
        total_size = 0
        df_index = 0
        if not os.path.exists(path):
            raise FileNotFoundError
        if os.path.isfile(path):
            dataframe = (
                self._dataframes[index] if self._dataframes is not None and len(self._dataframes) > index else None
            )
            file: File = self._build_file_from_path(path=path, dataframe=dataframe)
            total_size += file.size or 0
            files.append(file)
            df_index += 1
        else:
            dir_paths = glob.glob(f"{path}/**", recursive=True)
            for entry_path in filter(lambda x: os.path.isfile(x), sorted(dir_paths, key=str.lower)):
                new_index = index + df_index
                dataframe = (
                    self._dataframes[new_index]
                    if self._dataframes is not None and len(self._dataframes) > new_index
                    else None
                )
                glob_file: File = self._build_file_from_path(path=entry_path, dataframe=dataframe)
                total_size += glob_file.size or 0
                files.append(glob_file)
                df_index += 1
        return files, int(total_size), df_index

    def _build_file_from_path(self, path: str, dataframe: DataFrameType | None = None) -> File:
        entry_stats: os.stat_result = os.stat(path)
        name = path.split("\\")[-1]
        return File(
            name=name,
            size=entry_stats.st_size,
            fingerprint=self._compute_digest_for_path(path),
            updated_date=self._convert_date_to_iso(entry_stats.st_mtime),
            uri=f"file://{os.path.abspath(path)}",
            dataframe=dataframe,
        )

    @staticmethod
    def _compute_digest_for_path(path: str) -> str:
        sha = hashlib.sha256()
        bytes_array = bytearray(128 * 1024)
        mv = memoryview(bytes_array)
        with open(path, "rb", buffering=0) as file:
            n = file.readinto(mv)
            while n:
                sha.update(mv[:n])
                n = file.readinto(mv)
        return sha.hexdigest()

    @staticmethod
    def _convert_date_to_iso(timestamp: float) -> str:
        return datetime.datetime.fromtimestamp(timestamp).isoformat()



from __future__ import annotations

from vectice.api import json
from vectice.api.client import Client
from vectice.api.http_error_handlers import InvalidReferenceError, MissingReferenceError
from vectice.api.iteration import IterationApi
from vectice.api.last_assets import LastAssetApi
from vectice.api.phase import PhaseApi
from vectice.api.step import StepApi
from vectice.api.workspace import WorkspaceApi

__all__ = [
    "MissingReferenceError",
    "InvalidReferenceError",
    "Client",
    "WorkspaceApi",
    "json",
    "PhaseApi",
    "StepApi",
    "IterationApi",
    "LastAssetApi",
]



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from logging import Logger


def existing_dataset_logger(data: dict, dataset_name: str, _logger: Logger) -> str | None:
    """Identifies if the dataset and it's version exists already."""
    if data["useExistingDataset"]:
        existing_dataset_version = f"Dataset: {dataset_name}"
        if data["useExistingVersion"]:
            existing_dataset_version += f" with Version: {data['datasetVersion']['name']}"
        _logger.debug(f"{existing_dataset_version} already exists.")
        return existing_dataset_version
    return None


def existing_model_logger(data: dict, model_name: str, _logger: Logger):
    if data["useExistingModel"]:
        existing_model = f"Model: {model_name} already exists."
        _logger.debug(existing_model)
        return existing_model
    return



from __future__ import annotations


class CompatibilityOutput:
    def __init__(self, message: str, status: str, *args, **kwargs):
        self._message = message
        self._status = status

    @property
    def message(self):
        return self._message

    @property
    def status(self):
        return self._status



class DatasetOutput(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def name(self) -> str:
        return str(self["name"])

    @property
    def project_id(self) -> int:
        return int(self["projectId"])


class DatasetVersionOutput(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if "dataSet" in self:
            self._dataset: DatasetOutput = DatasetOutput(**self["dataSet"])
        else:
            self._dataset = None

    @property
    def id(self) -> str:
        return str(self["vecticeId"])

    @property
    def name(self) -> str:
        return str(self["name"])

    @property
    def dataset(self) -> DatasetOutput:
        return self._dataset



from __future__ import annotations

from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING

from typing_extensions import get_args

from vectice.models.resource.metadata.base import DatasetSourceOrigin, Metadata
from vectice.models.resource.metadata.dataframe_config import DataFrameType

if TYPE_CHECKING:
    from vectice.models.resource.metadata.base import DatasetSourceUsage


class Resource(metaclass=ABCMeta):
    """Base class for resources.

    Use Resource subclasses to assign datasets to steps.  The
    Vectice library supports a handful of common cases.  Additional
    cases are generally easy to supply by deriving from this base
    class.  In particular, subclasses must override this class'
    abstact methods (`_build_metadata()`, `_fetch_data()`).

    Examples:
        To create a custom resource class, inherit from `Resource`,
        and implement the `_build_metadata()` and `_fetch_data()` methods:

        ```python
        from vectice import Resource, DatasetSourceOrigin, FilesMetadata

        class MyResource(Resource):
            _origin = "Data source name"

            def __init__(
                self,
                paths: str | list[str],
            ):
                super().__init__(paths=paths)

            def _build_metadata(self) -> FilesMetadata:  # (1)
                files = ...  # fetch file list from your custom storage
                total_size = ...  # compute total file size, retrieve them from self._paths
                return FilesMetadata(
                    size=total_size,
                    origin=self._origin,
                    files=files,
                    usage=self.usage,
                )

            def _fetch_data(self) -> dict[str, bytes]:
                files_data = {}
                for file in self.metadata.files:
                    file_contents = ...  # fetch file contents from your custom storage
                    files_data[file.name] = file_contents
                return files_data
        ```

        1. Return [FilesMetadata][vectice.models.resource.metadata.FilesMetadata] for data stored in files,
            [DBMetadata][vectice.models.resource.metadata.DBMetadata] for data stored in a database.
    """

    # origin must be a string because users can subclass resources
    # and set a custom origin
    _origin: str = DatasetSourceOrigin.OTHER.value

    _files_limit = 5000

    @abstractmethod
    def __init__(self, paths: list[str] | str, dataframes: DataFrameType | list[DataFrameType] | None = None):
        """Initialize a resource."""
        self._metadata: Metadata | None = None
        self._data: dict | None = None

        self._paths = paths if isinstance(paths, list) else [paths]

        self._dataframes = dataframes if isinstance(dataframes, list) or dataframes is None else [dataframes]
        if self._dataframes is not None:
            for dataframe in self._dataframes:
                if dataframe is not None and not isinstance(dataframe, get_args(DataFrameType)):
                    raise ValueError(
                        f"Argument 'dataframe' of type '{type(dataframe)}' is invalid, only Pandas DataFrame is supported."
                    )

    @property
    def data(self) -> dict:
        """The resource's data.

        Returns:
            The resource's data.
        """
        if self._data is None:
            self._data = self._fetch_data()
        return self._data

    @abstractmethod
    def _fetch_data(
        self,
    ) -> dict:
        pass

    @abstractmethod
    def _build_metadata(self) -> Metadata:
        pass

    @property
    def metadata(self) -> Metadata:
        """The resource's metadata.

        Returns:
            The resource's metadata.
        """
        if self._metadata is None:
            self._metadata = self._build_metadata()
        return self._metadata

    @metadata.setter
    def metadata(self, value):
        """Set the resource's metadata.

        Parameters:
            value: The metadata to set.
        """
        self._metadata = value

    @property
    def usage(self) -> DatasetSourceUsage | None:
        """The resource's usage.

        Returns:
            The resource's usage.
        """
        return self.metadata.usage

    @usage.setter
    def usage(self, value: DatasetSourceUsage):
        """Set the resource's usage.

        Parameters:
            value: The usage to set.
        """
        self.metadata.usage = value



from typing import Union

from pandas.core.frame import DataFrame as PandasDF
from pyspark.pandas.frame import DataFrame as PysparkPandasDF
from pyspark.sql.dataframe import DataFrame as SparkDF

from vectice.models.resource.metadata.df_wrapper_resource import DataFrameWrapper

DataFramePandasType = Union[PysparkPandasDF, PandasDF]
DataFrameTypeWithoutWrapper = Union[DataFramePandasType, SparkDF]
DataFrameType = Union[DataFrameTypeWithoutWrapper, DataFrameWrapper]

MIN_ROWS_CAPTURE_STATS = 100
MAX_COLUMNS_CAPTURE_STATS = 100
